import asyncio
from typing import cast, Any, Awaitable, Optional

from pytest import mark, raises

from graphql.error import GraphQLError
from graphql.execution import execute, execute_sync
from graphql.language import parse, FieldNode, OperationDefinitionNode
from graphql.pyutils import inspect, Undefined
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLField,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLScalarType,
    GraphQLString,
    GraphQLUnionType,
    ResponsePath,
)


def describe_execute_handles_basic_execution_tasks():
    # noinspection PyTypeChecker
    def throws_if_no_document_is_provided():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        with raises(TypeError) as exc_info:
            assert execute_sync(schema=schema, document=None)  # type: ignore

        assert str(exc_info.value) == "Must provide document."

    # noinspection PyTypeChecker
    def throws_if_no_schema_is_provided():
        document = parse("{ field }")

        with raises(TypeError) as exc_info:
            assert execute_sync(schema=None, document=document)  # type: ignore

        assert str(exc_info.value) == "Expected None to be a GraphQL schema."

    def throws_on_invalid_variables():
        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type",
                {
                    "fieldA": GraphQLField(
                        GraphQLString, args={"argA": GraphQLArgument(GraphQLInt)}
                    )
                },
            )
        )
        document = parse(
            """
            query ($a: Int) {
              fieldA(argA: $a)
            }
            """
        )
        variable_values = "{'a': 1}"

        with raises(TypeError) as exc_info:
            assert execute_sync(
                schema=schema,
                document=document,
                variable_values=variable_values,  # type: ignore
            )

        assert str(exc_info.value) == (
            "Variable values must be provided as a dictionary"
            " with variable names as keys. Perhaps look to see"
            " if an unparsed JSON string was provided."
        )

    def accepts_positional_arguments():
        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type",
                {"a": GraphQLField(GraphQLString, resolve=lambda obj, *args: obj)},
            )
        )

        result = execute_sync(schema, parse("{ a }"), "rootValue")

        assert result == ({"a": "rootValue"}, None)

    @mark.asyncio
    async def executes_arbitrary_code():
        # noinspection PyMethodMayBeStatic,PyMethodMayBeStatic
        class Data:
            def a(self, _info):
                return "Apple"

            def b(self, _info):
                return "Banana"

            def c(self, _info):
                return "Cookie"

            def d(self, _info):
                return "Donut"

            def e(self, _info):
                return "Egg"

            f = "Fish"

            # Called only by DataType::pic static resolver
            def pic(self, _info, size: int):
                return f"Pic of size: {size}"

            def deep(self, _info):
                return DeepData()

            def promise(self, _info):
                return promise_data()

        # noinspection PyMethodMayBeStatic,PyMethodMayBeStatic
        class DeepData:
            def a(self, _info):
                return "Already Been Done"

            def b(self, _info):
                return "Boring"

            def c(self, _info):
                return ["Contrived", None, "Confusing"]

            def deeper(self, _info):
                return [Data(), None, Data()]

        async def promise_data():
            await asyncio.sleep(0)
            return Data()

        DeepDataType: GraphQLObjectType

        DataType = GraphQLObjectType(
            "DataType",
            lambda: {
                "a": GraphQLField(GraphQLString),
                "b": GraphQLField(GraphQLString),
                "c": GraphQLField(GraphQLString),
                "d": GraphQLField(GraphQLString),
                "e": GraphQLField(GraphQLString),
                "f": GraphQLField(GraphQLString),
                "pic": GraphQLField(
                    GraphQLString,
                    args={"size": GraphQLArgument(GraphQLInt)},
                    resolve=lambda obj, info, size: obj.pic(info, size),
                ),
                "deep": GraphQLField(DeepDataType),
                "promise": GraphQLField(DataType),
            },
        )

        DeepDataType = GraphQLObjectType(
            "DeepDataType",
            {
                "a": GraphQLField(GraphQLString),
                "b": GraphQLField(GraphQLString),
                "c": GraphQLField(GraphQLList(GraphQLString)),
                "deeper": GraphQLField(GraphQLList(DataType)),
            },
        )

        document = parse(
            """
            query ($size: Int) {
              a,
              b,
              x: c
              ...c
              f
              ...on DataType {
                pic(size: $size)
                promise {
                  a
                }
              }
              deep {
                a
                b
                c
                deeper {
                  a
                  b
                }
              }
            }

            fragment c on DataType {
              d
              e
            }
            """
        )

        awaitable_result = execute(
            GraphQLSchema(DataType), document, Data(), variable_values={"size": 100}
        )
        assert isinstance(awaitable_result, Awaitable)
        result = await awaitable_result

        assert result == (
            {
                "a": "Apple",
                "b": "Banana",
                "x": "Cookie",
                "d": "Donut",
                "e": "Egg",
                "f": "Fish",
                "pic": "Pic of size: 100",
                "promise": {"a": "Apple"},
                "deep": {
                    "a": "Already Been Done",
                    "b": "Boring",
                    "c": ["Contrived", None, "Confusing"],
                    "deeper": [
                        {"a": "Apple", "b": "Banana"},
                        None,
                        {"a": "Apple", "b": "Banana"},
                    ],
                },
            },
            None,
        )

    def merges_parallel_fragments():
        Type = GraphQLObjectType(
            "Type",
            lambda: {
                "a": GraphQLField(GraphQLString, resolve=lambda *_args: "Apple"),
                "b": GraphQLField(GraphQLString, resolve=lambda *_args: "Banana"),
                "c": GraphQLField(GraphQLString, resolve=lambda *_args: "Cherry"),
                "deep": GraphQLField(Type, resolve=lambda *_args: {}),
            },
        )
        schema = GraphQLSchema(Type)

        ast = parse(
            """
            { a, ...FragOne, ...FragTwo }

            fragment FragOne on Type {
              b
              deep { b, deeper: deep { b } }
            }

            fragment FragTwo on Type {
              c
              deep { c, deeper: deep { c } }
            }
            """
        )

        result = execute_sync(schema, ast)
        assert result == (
            {
                "a": "Apple",
                "b": "Banana",
                "c": "Cherry",
                "deep": {
                    "b": "Banana",
                    "c": "Cherry",
                    "deeper": {"b": "Banana", "c": "Cherry"},
                },
            },
            None,
        )

    def provides_info_about_current_execution_state():
        resolved_infos = []

        def resolve(_obj, info):
            resolved_infos.append(info)

        test_type = GraphQLObjectType(
            "Test", {"test": GraphQLField(GraphQLString, resolve=resolve)}
        )

        schema = GraphQLSchema(test_type)

        document = parse("query ($var: String) { result: test }")
        root_value = {"root": "val"}
        variable_values = {"var": "abc"}
        execute_sync(schema, document, root_value, variable_values=variable_values)

        assert len(resolved_infos) == 1
        operation = cast(OperationDefinitionNode, document.definitions[0])
        assert operation and operation.kind == "operation_definition"
        field = cast(FieldNode, operation.selection_set.selections[0])

        assert resolved_infos[0] == GraphQLResolveInfo(
            field_name="test",
            field_nodes=[field],
            return_type=GraphQLString,
            parent_type=cast(GraphQLObjectType, schema.query_type),
            path=ResponsePath(None, "result", "Test"),
            schema=schema,
            fragments={},
            root_value=root_value,
            operation=operation,
            variable_values=variable_values,
            context=None,
            is_awaitable=resolved_infos[0].is_awaitable,
        )

    def it_populates_path_correctly_with_complex_types():
        path: Optional[ResponsePath] = None

        def resolve(_val, info):
            nonlocal path
            path = info.path

        def resolve_type(_val, _info, _type):
            return "SomeObject"

        some_object = GraphQLObjectType(
            "SomeObject", {"test": GraphQLField(GraphQLString, resolve=resolve)}
        )
        some_union = GraphQLUnionType(
            "SomeUnion", [some_object], resolve_type=resolve_type
        )
        test_type = GraphQLObjectType(
            "SomeQuery",
            {
                "test": GraphQLField(
                    GraphQLNonNull(GraphQLList(GraphQLNonNull(some_union)))
                )
            },
        )
        schema = GraphQLSchema(test_type)
        root_value: Any = {"test": [{}]}
        document = parse(
            """
            query {
              l1: test {
                ... on SomeObject {
                  l2: test
                }
              }
            }
            """
        )

        execute_sync(schema, document, root_value)

        assert path is not None
        prev, key, typename = path
        assert key == "l2"
        assert typename == "SomeObject"
        prev, key, typename = prev
        assert key == 0
        assert typename is None
        prev, key, typename = prev
        assert key == "l1"
        assert typename == "SomeQuery"
        assert prev is None

    def threads_root_value_context_correctly():
        resolved_values = []

        class Data:
            context_thing = "thing"

        def resolve(obj, _info):
            resolved_values.append(obj)

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type", {"a": GraphQLField(GraphQLString, resolve=resolve)}
            )
        )

        document = parse("query Example { a }")
        root_value = Data()
        execute_sync(schema, document, root_value)

        assert len(resolved_values) == 1
        assert resolved_values[0] is root_value

    def correctly_threads_arguments():
        resolved_args = []

        def resolve(_obj, _info, **args):
            resolved_args.append(args)

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type",
                {
                    "b": GraphQLField(
                        GraphQLString,
                        args={
                            "numArg": GraphQLArgument(GraphQLInt),
                            "stringArg": GraphQLArgument(GraphQLString),
                        },
                        resolve=resolve,
                    )
                },
            )
        )

        document = parse(
            """
            query Example {
              b(numArg: 123, stringArg: "foo")
            }
            """
        )

        execute_sync(schema, document)

        assert len(resolved_args) == 1
        assert resolved_args[0] == {"numArg": 123, "stringArg": "foo"}

    @mark.asyncio
    async def nulls_out_error_subtrees():
        document = parse(
            """
            {
              syncOk
              syncError
              syncRawError
              syncReturnError
              syncReturnErrorList
              asyncOk
              asyncError
              asyncRawError
              asyncReturnError
              asyncReturnErrorWithExtensions
            }
            """
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type",
                {
                    "syncOk": GraphQLField(GraphQLString),
                    "syncError": GraphQLField(GraphQLString),
                    "syncRawError": GraphQLField(GraphQLString),
                    "syncReturnError": GraphQLField(GraphQLString),
                    "syncReturnErrorList": GraphQLField(GraphQLList(GraphQLString)),
                    "asyncOk": GraphQLField(GraphQLString),
                    "asyncError": GraphQLField(GraphQLString),
                    "asyncErrorWithExtensions": GraphQLField(GraphQLString),
                    "asyncRawError": GraphQLField(GraphQLString),
                    "asyncReturnError": GraphQLField(GraphQLString),
                    "asyncReturnErrorWithExtensions": GraphQLField(GraphQLString),
                },
            )
        )

        # noinspection PyPep8Naming,PyMethodMayBeStatic
        class Data:
            def syncOk(self, _info):
                return "sync ok"

            def syncError(self, _info):
                raise GraphQLError("Error getting syncError")

            def syncRawError(self, _info):
                raise Exception("Error getting syncRawError")

            def syncReturnError(self, _info):
                return Exception("Error getting syncReturnError")

            def syncReturnErrorList(self, _info):
                return [
                    "sync0",
                    Exception("Error getting syncReturnErrorList1"),
                    "sync2",
                    Exception("Error getting syncReturnErrorList3"),
                ]

            async def asyncOk(self, _info):
                return "async ok"

            async def asyncError(self, _info):
                raise GraphQLError("Error getting asyncError")

            async def asyncRawError(self, _info):
                raise Exception("Error getting asyncRawError")

            async def asyncReturnError(self, _info):
                return GraphQLError("Error getting asyncReturnError")

            async def asyncReturnErrorWithExtensions(self, _info):
                return GraphQLError(
                    "Error getting asyncReturnErrorWithExtensions",
                    extensions={"foo": "bar"},
                )

        awaitable_result = execute(schema, document, Data())
        assert isinstance(awaitable_result, Awaitable)
        result = await awaitable_result

        assert result == (
            {
                "syncOk": "sync ok",
                "syncError": None,
                "syncRawError": None,
                "syncReturnError": None,
                "syncReturnErrorList": ["sync0", None, "sync2", None],
                "asyncOk": "async ok",
                "asyncError": None,
                "asyncRawError": None,
                "asyncReturnError": None,
                "asyncReturnErrorWithExtensions": None,
            },
            [
                {
                    "message": "Error getting syncError",
                    "locations": [(4, 15)],
                    "path": ["syncError"],
                },
                {
                    "message": "Error getting syncRawError",
                    "locations": [(5, 15)],
                    "path": ["syncRawError"],
                },
                {
                    "message": "Error getting syncReturnError",
                    "locations": [(6, 15)],
                    "path": ["syncReturnError"],
                },
                {
                    "message": "Error getting syncReturnErrorList1",
                    "locations": [(7, 15)],
                    "path": ["syncReturnErrorList", 1],
                },
                {
                    "message": "Error getting syncReturnErrorList3",
                    "locations": [(7, 15)],
                    "path": ["syncReturnErrorList", 3],
                },
                {
                    "message": "Error getting asyncError",
                    "locations": [(9, 15)],
                    "path": ["asyncError"],
                },
                {
                    "message": "Error getting asyncRawError",
                    "locations": [(10, 15)],
                    "path": ["asyncRawError"],
                },
                {
                    "message": "Error getting asyncReturnError",
                    "locations": [(11, 15)],
                    "path": ["asyncReturnError"],
                },
                {
                    "message": "Error getting asyncReturnErrorWithExtensions",
                    "locations": [(12, 15)],
                    "path": ["asyncReturnErrorWithExtensions"],
                    "extensions": {"foo": "bar"},
                },
            ],
        )

    def full_response_path_is_included_for_non_nullable_fields():
        def resolve_ok(*_args):
            return {}

        def resolve_error(*_args):
            raise Exception("Catch me if you can")

        A = GraphQLObjectType(
            "A",
            lambda: {
                "nullableA": GraphQLField(A, resolve=resolve_ok),
                "nonNullA": GraphQLField(GraphQLNonNull(A), resolve=resolve_ok),
                "throws": GraphQLField(GraphQLNonNull(A), resolve=resolve_error),
            },
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "query", lambda: {"nullableA": GraphQLField(A, resolve=resolve_ok)}
            )
        )

        document = parse(
            """
            query {
              nullableA {
                aliasedA: nullableA {
                  nonNullA {
                    anotherA: nonNullA {
                      throws
                    }
                  }
                }
              }
            }
            """
        )

        assert execute_sync(schema, document) == (
            {"nullableA": {"aliasedA": None}},
            [
                {
                    "message": "Catch me if you can",
                    "locations": [(7, 23)],
                    "path": ["nullableA", "aliasedA", "nonNullA", "anotherA", "throws"],
                }
            ],
        )

    def uses_the_inline_operation_if_no_operation_name_is_provided():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse("{ a }")

        class Data:
            a = "b"

        result = execute_sync(schema, document, Data())
        assert result == ({"a": "b"}, None)

    def uses_the_only_operation_if_no_operation_name_is_provided():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse("query Example { a }")

        class Data:
            a = "b"

        result = execute_sync(schema, document, Data())
        assert result == ({"a": "b"}, None)

    def uses_the_named_operation_if_operation_name_is_provided():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse(
            """
            query Example { first: a }
            query OtherExample { second: a }
            """
        )

        class Data:
            a = "b"

        result = execute_sync(schema, document, Data(), operation_name="OtherExample")
        assert result == ({"second": "b"}, None)

    def provides_error_if_no_operation_is_provided():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse("fragment Example on Type { a }")

        class Data:
            a = "b"

        result = execute_sync(schema, document, Data())
        assert result == (None, [{"message": "Must provide an operation."}])

    def errors_if_no_operation_name_is_provided_with_multiple_operations():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse(
            """
            query Example { a }
            query OtherExample { a }
            """
        )

        result = execute_sync(schema, document)
        assert result == (
            None,
            [
                {
                    "message": "Must provide operation name if query contains"
                    " multiple operations."
                }
            ],
        )

    def errors_if_unknown_operation_name_is_provided():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse(
            """
            query Example { a }
            query OtherExample { a }
            """
        )

        result = execute_sync(schema, document, operation_name="UnknownExample")
        assert result == (
            None,
            [{"message": "Unknown operation named 'UnknownExample'."}],
        )

    def errors_if_empty_string_is_provided_as_operation_name():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse("{ a }")

        result = execute_sync(schema, document, operation_name="")
        assert result == (
            None,
            [{"message": "Unknown operation named ''."}],
        )

    def uses_the_query_schema_for_queries():
        schema = GraphQLSchema(
            GraphQLObjectType("Q", {"a": GraphQLField(GraphQLString)}),
            GraphQLObjectType("M", {"c": GraphQLField(GraphQLString)}),
            GraphQLObjectType("S", {"a": GraphQLField(GraphQLString)}),
        )

        document = parse(
            """
            query Q { a }
            mutation M { c }
            subscription S { a }
            """
        )

        class Data:
            a = "b"
            c = "d"

        result = execute_sync(schema, document, Data(), operation_name="Q")
        assert result == ({"a": "b"}, None)

    def uses_the_mutation_schema_for_mutations():
        schema = GraphQLSchema(
            GraphQLObjectType("Q", {"a": GraphQLField(GraphQLString)}),
            GraphQLObjectType("M", {"c": GraphQLField(GraphQLString)}),
        )

        document = parse(
            """
            query Q { a }
            mutation M { c }
            """
        )

        class Data:
            a = "b"
            c = "d"

        result = execute_sync(schema, document, Data(), operation_name="M")
        assert result == ({"c": "d"}, None)

    def uses_the_subscription_schema_for_subscriptions():
        schema = GraphQLSchema(
            query=GraphQLObjectType("Q", {"a": GraphQLField(GraphQLString)}),
            subscription=GraphQLObjectType("S", {"a": GraphQLField(GraphQLString)}),
        )

        document = parse(
            """
            query Q { a }
            subscription S { a }
            """
        )

        class Data:
            a = "b"
            c = "d"

        result = execute_sync(schema, document, Data(), operation_name="S")
        assert result == ({"a": "b"}, None)

    def resolves_to_an_error_if_schema_does_not_support_operation():
        schema = GraphQLSchema(assume_valid=True)

        document = parse(
            """
            query Q { __typename }
            mutation M { __typename }
            subscription S { __typename }
            """
        )

        assert execute_sync(schema, document, operation_name="Q") == (
            None,
            [
                {
                    "message": "Schema is not configured to execute query operation.",
                    "locations": [(2, 13)],
                }
            ],
        )

        assert execute_sync(schema, document, operation_name="M") == (
            None,
            [
                {
                    "message": "Schema is not configured to execute"
                    " mutation operation.",
                    "locations": [(3, 13)],
                }
            ],
        )

        assert execute_sync(schema, document, operation_name="S") == (
            None,
            [
                {
                    "message": "Schema is not configured to execute"
                    " subscription operation.",
                    "locations": [(4, 13)],
                }
            ],
        )

    @mark.asyncio
    async def correct_field_ordering_despite_execution_order():
        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type",
                {
                    "a": GraphQLField(GraphQLString),
                    "b": GraphQLField(GraphQLString),
                    "c": GraphQLField(GraphQLString),
                    "d": GraphQLField(GraphQLString),
                    "e": GraphQLField(GraphQLString),
                },
            )
        )

        document = parse("{ a, b, c, d, e}")

        # noinspection PyMethodMayBeStatic,PyMethodMayBeStatic
        class Data:
            def a(self, _info):
                return "a"

            async def b(self, _info):
                return "b"

            def c(self, _info):
                return "c"

            async def d(self, _info):
                return "d"

            def e(self, _info):
                return "e"

        awaitable_result = execute(schema, document, Data())
        assert isinstance(awaitable_result, Awaitable)
        result = await awaitable_result

        assert result == ({"a": "a", "b": "b", "c": "c", "d": "d", "e": "e"}, None)

    def avoids_recursion():
        schema = GraphQLSchema(
            GraphQLObjectType("Type", {"a": GraphQLField(GraphQLString)})
        )

        document = parse(
            """
            query Q {
              a
              ...Frag
              ...Frag
            }

            fragment Frag on Type {
              a,
              ...Frag
            }
            """
        )

        class Data:
            a = "b"

        result = execute_sync(schema, document, Data(), operation_name="Q")

        assert result == ({"a": "b"}, None)

    def ignores_missing_sub_selections_on_fields():
        some_type = GraphQLObjectType("SomeType", {"b": GraphQLField(GraphQLString)})
        schema = GraphQLSchema(
            GraphQLObjectType("Query", {"a": GraphQLField(some_type)})
        )
        document = parse("{ a }")
        root_value = {"a": {"b": "c"}}

        result = execute_sync(schema, document, root_value)
        assert result == ({"a": {}}, None)

    def does_not_include_illegal_fields_in_output():
        schema = GraphQLSchema(
            GraphQLObjectType("Q", {"a": GraphQLField(GraphQLString)})
        )

        document = parse("{ thisIsIllegalDoNotIncludeMe }")

        result = execute_sync(schema, document)

        assert result == ({}, None)

    def does_not_include_arguments_that_were_not_set():
        schema = GraphQLSchema(
            GraphQLObjectType(
                "Type",
                {
                    "field": GraphQLField(
                        GraphQLString,
                        args={
                            "a": GraphQLArgument(GraphQLBoolean),
                            "b": GraphQLArgument(GraphQLBoolean),
                            "c": GraphQLArgument(GraphQLBoolean),
                            "d": GraphQLArgument(GraphQLInt),
                            "e": GraphQLArgument(GraphQLInt),
                        },
                        resolve=lambda _source, _info, **args: inspect(args),
                    )
                },
            )
        )

        document = parse("{ field(a: true, c: false, e: 0) }")

        assert execute_sync(schema, document) == (
            {"field": "{'a': True, 'c': False, 'e': 0}"},
            None,
        )

    @mark.asyncio
    async def fails_when_is_type_of_check_is_not_met():
        class Special:
            value: str

            def __init__(self, value):
                self.value = value

        class NotSpecial:
            value: str

            def __init__(self, value):
                self.value = value

        def is_type_of_special(obj, _info):
            is_special = isinstance(obj, Special)
            if not _info.context["async"]:
                return is_special

            async def async_is_special():
                return is_special

            return async_is_special()

        SpecialType = GraphQLObjectType(
            "SpecialType",
            {"value": GraphQLField(GraphQLString)},
            is_type_of=is_type_of_special,
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query", {"specials": GraphQLField(GraphQLList(SpecialType))}
            )
        )

        document = parse("{ specials { value } }")
        root_value = {"specials": [Special("foo"), NotSpecial("bar")]}

        result = execute_sync(schema, document, root_value, {"async": False})
        assert not isinstance(result, Awaitable)
        assert result == (
            {"specials": [{"value": "foo"}, None]},
            [
                {
                    "message": "Expected value of type 'SpecialType' but got:"
                    " <NotSpecial instance>.",
                    "locations": [(1, 3)],
                    "path": ["specials", 1],
                }
            ],
        )

        async_result = execute(schema, document, root_value, {"async": True})
        assert isinstance(async_result, Awaitable)
        awaited_result = await async_result
        assert awaited_result == result

    def fails_when_serialize_of_custom_scalar_does_not_return_a_value():
        custom_scalar = GraphQLScalarType(
            "CustomScalar", serialize=lambda _value: Undefined  # returns nothing
        )
        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "customScalar": GraphQLField(
                        custom_scalar, resolve=lambda *_args: "CUSTOM_VALUE"
                    )
                },
            )
        )

        result = execute_sync(schema, parse("{ customScalar }"))
        assert result == (
            {"customScalar": None},
            [
                {
                    "message": "Expected `CustomScalar.serialize('CUSTOM_VALUE')`"
                    " to return non-nullable value, returned: Undefined",
                    "locations": [(1, 3)],
                    "path": ["customScalar"],
                }
            ],
        )

    def executes_ignoring_invalid_non_executable_definitions():
        schema = GraphQLSchema(
            GraphQLObjectType("Query", {"foo": GraphQLField(GraphQLString)})
        )

        document = parse(
            """
            { foo }

            type Query { bar: String }
            """
        )

        result = execute_sync(schema, document)
        assert result == ({"foo": None}, None)

    def uses_a_custom_field_resolver():
        schema = GraphQLSchema(
            GraphQLObjectType("Query", {"foo": GraphQLField(GraphQLString)})
        )
        document = parse("{ foo }")

        def field_resolver(_source, info):
            # For the purposes of test, just return the name of the field!
            return info.field_name

        result = execute_sync(schema, document, field_resolver=field_resolver)
        assert result == ({"foo": "foo"}, None)

    def uses_a_custom_type_resolver():
        document = parse("{ foo { bar } }")

        foo_interface = GraphQLInterfaceType(
            "FooInterface", {"bar": GraphQLField(GraphQLString)}
        )

        foo_object = GraphQLObjectType(
            "FooObject", {"bar": GraphQLField(GraphQLString)}, [foo_interface]
        )

        schema = GraphQLSchema(
            GraphQLObjectType("Query", {"foo": GraphQLField(foo_interface)}),
            types=[foo_object],
        )

        possible_types = None

        def type_resolver(_source, info, abstract_type):
            # Resolver should be able to figure out all possible types on its own
            nonlocal possible_types
            possible_types = info.schema.get_possible_types(abstract_type)
            return "FooObject"

        root_value = {"foo": {"bar": "bar"}}
        result = execute_sync(schema, document, root_value, type_resolver=type_resolver)

        assert result == ({"foo": {"bar": "bar"}}, None)
        assert possible_types == [foo_object]



import sys
from asyncio import CancelledError, Event, ensure_future, sleep

from pytest import mark, raises

from graphql.execution import MapAsyncIterator


try:  # pragma: no cover
    anext
except NameError:  # pragma: no cover (Python < 3.10)
    # noinspection PyShadowingBuiltins
    async def anext(iterator):
        """Return the next item from an async iterator."""
        return await iterator.__anext__()


def describe_map_async_iterator():
    @mark.asyncio
    async def maps_over_async_generator():
        async def source():
            yield 1
            yield 2
            yield 3

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4
        assert await anext(doubles) == 6
        with raises(StopAsyncIteration):
            assert await anext(doubles)

    @mark.asyncio
    async def maps_over_async_iterable():
        items = [1, 2, 3]

        class Iterable:
            def __aiter__(self):
                return self

            async def __anext__(self):
                try:
                    return items.pop(0)
                except IndexError:
                    raise StopAsyncIteration

        doubles = MapAsyncIterator(Iterable(), lambda x: x + x)

        values = [value async for value in doubles]

        assert not items
        assert values == [2, 4, 6]

    @mark.asyncio
    async def compatible_with_async_for():
        async def source():
            yield 1
            yield 2
            yield 3

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        values = [value async for value in doubles]

        assert values == [2, 4, 6]

    @mark.asyncio
    async def maps_over_async_values_with_async_function():
        async def source():
            yield 1
            yield 2
            yield 3

        async def double(x):
            return x + x

        doubles = MapAsyncIterator(source(), double)

        values = [value async for value in doubles]

        assert values == [2, 4, 6]

    @mark.asyncio
    async def allows_returning_early_from_mapped_async_generator():
        async def source():
            yield 1
            yield 2
            yield 3  # pragma: no cover

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4

        # Early return
        await doubles.aclose()

        # Subsequent next calls
        with raises(StopAsyncIteration):
            await anext(doubles)
        with raises(StopAsyncIteration):
            await anext(doubles)

    @mark.asyncio
    async def allows_returning_early_from_mapped_async_iterable():
        items = [1, 2, 3]

        class Iterable:
            def __aiter__(self):
                return self

            async def __anext__(self):
                try:
                    return items.pop(0)
                except IndexError:  # pragma: no cover
                    raise StopAsyncIteration

        doubles = MapAsyncIterator(Iterable(), lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4

        # Early return
        await doubles.aclose()

        # Subsequent next calls
        with raises(StopAsyncIteration):
            await anext(doubles)
        with raises(StopAsyncIteration):
            await anext(doubles)

    @mark.asyncio
    async def passes_through_early_return_from_async_values():
        async def source():
            try:
                yield 1
                yield 2
                yield 3  # pragma: no cover
            finally:
                yield "Done"
                yield "Last"

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4

        # Early return
        await doubles.aclose()

        # Subsequent next calls may yield from finally block
        assert await anext(doubles) == "LastLast"
        with raises(GeneratorExit):
            assert await anext(doubles)

    @mark.asyncio
    async def allows_throwing_errors_through_async_iterable():
        items = [1, 2, 3]

        class Iterable:
            def __aiter__(self):
                return self

            async def __anext__(self):
                try:
                    return items.pop(0)
                except IndexError:  # pragma: no cover
                    raise StopAsyncIteration

        doubles = MapAsyncIterator(Iterable(), lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4

        # Throw error
        with raises(RuntimeError, match="Ouch") as exc_info:
            await doubles.athrow(RuntimeError("Ouch"))

        assert str(exc_info.value) == "Ouch"

        with raises(StopAsyncIteration):
            await anext(doubles)
        with raises(StopAsyncIteration):
            await anext(doubles)

    @mark.asyncio
    async def allows_throwing_errors_with_values_through_async_iterators():
        class Iterator:
            def __aiter__(self):
                return self

            async def __anext__(self):
                return 1

        one = MapAsyncIterator(Iterator(), lambda x: x)

        assert await anext(one) == 1

        # Throw error with value passed separately
        try:
            raise RuntimeError("Ouch")
        except RuntimeError as error:
            with raises(RuntimeError, match="Ouch") as exc_info:
                await one.athrow(error.__class__, error)

            assert exc_info.value is error
            assert exc_info.tb is error.__traceback__

        with raises(StopAsyncIteration):
            await anext(one)

    @mark.asyncio
    async def allows_throwing_errors_with_traceback_through_async_iterators():
        class Iterator:
            def __aiter__(self):
                return self

            async def __anext__(self):
                return 1

        one = MapAsyncIterator(Iterator(), lambda x: x)

        assert await anext(one) == 1

        # Throw error with traceback passed separately
        try:
            raise RuntimeError("Ouch")
        except RuntimeError as error:
            with raises(RuntimeError) as exc_info:
                await one.athrow(error.__class__, None, error.__traceback__)

            assert exc_info.tb and error.__traceback__
            assert exc_info.tb.tb_frame is error.__traceback__.tb_frame

        with raises(StopAsyncIteration):
            await anext(one)

    @mark.asyncio
    async def passes_through_caught_errors_through_async_generators():
        async def source():
            try:
                yield 1
                yield 2
                yield 3  # pragma: no cover
            except Exception as e:
                yield e

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4

        # Throw error
        await doubles.athrow(RuntimeError("ouch"))

        with raises(StopAsyncIteration):
            await anext(doubles)
        with raises(StopAsyncIteration):
            await anext(doubles)

    @mark.asyncio
    async def does_not_normally_map_over_thrown_errors():
        async def source():
            yield "Hello"
            raise RuntimeError("Goodbye")

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        assert await anext(doubles) == "HelloHello"

        with raises(RuntimeError) as exc_info:
            await anext(doubles)

        assert str(exc_info.value) == "Goodbye"

    @mark.asyncio
    async def does_not_normally_map_over_externally_thrown_errors():
        async def source():
            yield "Hello"

        doubles = MapAsyncIterator(source(), lambda x: x + x)

        assert await anext(doubles) == "HelloHello"

        with raises(RuntimeError) as exc_info:
            await doubles.athrow(RuntimeError("Goodbye"))

        assert str(exc_info.value) == "Goodbye"

    @mark.asyncio
    async def can_use_simple_iterator_instead_of_generator():
        async def source():
            yield 1
            yield 2
            yield 3

        class Source:
            def __init__(self):
                self.counter = 0

            def __aiter__(self):
                return self

            async def __anext__(self):
                self.counter += 1
                if self.counter > 3:
                    raise StopAsyncIteration
                return self.counter

        def double(x):
            return x + x

        for iterator in source, Source:
            doubles = MapAsyncIterator(iterator(), double)

            await doubles.aclose()

            with raises(StopAsyncIteration):
                await anext(doubles)

            doubles = MapAsyncIterator(iterator(), double)

            assert await anext(doubles) == 2
            assert await anext(doubles) == 4
            assert await anext(doubles) == 6

            with raises(StopAsyncIteration):
                await anext(doubles)

            doubles = MapAsyncIterator(iterator(), double)

            assert await anext(doubles) == 2
            assert await anext(doubles) == 4

            # Throw error
            with raises(RuntimeError) as exc_info:
                await doubles.athrow(RuntimeError("ouch"))

            assert str(exc_info.value) == "ouch"

            with raises(StopAsyncIteration):
                await anext(doubles)
            with raises(StopAsyncIteration):
                await anext(doubles)

            await doubles.athrow(RuntimeError("no more ouch"))

            with raises(StopAsyncIteration):
                await anext(doubles)

            await doubles.aclose()

            doubles = MapAsyncIterator(iterator(), double)

            assert await anext(doubles) == 2
            assert await anext(doubles) == 4

            try:
                raise ValueError("bad")
            except ValueError:
                tb = sys.exc_info()[2]

            # Throw error
            with raises(ValueError):
                await doubles.athrow(ValueError, None, tb)

        await sleep(0)

    @mark.asyncio
    async def stops_async_iteration_on_close():
        async def source():
            yield 1
            await Event().wait()  # Block forever
            yield 2  # pragma: no cover
            yield 3  # pragma: no cover

        singles = source()
        doubles = MapAsyncIterator(singles, lambda x: x * 2)

        result = await anext(doubles)
        assert result == 2

        # Make sure it is blocked
        doubles_future = ensure_future(anext(doubles))
        await sleep(0.05)
        assert not doubles_future.done()

        # Unblock and watch StopAsyncIteration propagate
        await doubles.aclose()
        await sleep(0.05)
        assert doubles_future.done()
        assert isinstance(doubles_future.exception(), StopAsyncIteration)

        with raises(StopAsyncIteration):
            await anext(singles)

    @mark.asyncio
    async def can_unset_closed_state_of_async_iterator():
        items = [1, 2, 3]

        class Iterator:
            def __init__(self):
                self.is_closed = False

            def __aiter__(self):
                return self

            async def __anext__(self):
                if self.is_closed:
                    raise StopAsyncIteration
                try:
                    return items.pop(0)
                except IndexError:
                    raise StopAsyncIteration

            async def aclose(self):
                self.is_closed = True

        iterator = Iterator()
        doubles = MapAsyncIterator(iterator, lambda x: x + x)

        assert await anext(doubles) == 2
        assert await anext(doubles) == 4
        assert not iterator.is_closed
        await doubles.aclose()
        assert iterator.is_closed
        with raises(StopAsyncIteration):
            await anext(iterator)
        with raises(StopAsyncIteration):
            await anext(doubles)
        assert doubles.is_closed

        iterator.is_closed = False
        doubles.is_closed = False
        assert not doubles.is_closed

        assert await anext(doubles) == 6
        assert not doubles.is_closed
        assert not iterator.is_closed
        with raises(StopAsyncIteration):
            await anext(iterator)
        with raises(StopAsyncIteration):
            await anext(doubles)
        assert not doubles.is_closed
        assert not iterator.is_closed

    @mark.asyncio
    async def can_cancel_async_iterator_while_waiting():
        class Iterator:
            def __init__(self):
                self.is_closed = False
                self.value = 1

            def __aiter__(self):
                return self

            async def __anext__(self):
                try:
                    await sleep(0.5)
                    return self.value  # pragma: no cover
                except CancelledError:
                    self.value = -1
                    raise

            async def aclose(self):
                self.is_closed = True

        iterator = Iterator()
        doubles = MapAsyncIterator(iterator, lambda x: x + x)  # pragma: no cover exit
        cancelled = False

        async def iterator_task():
            nonlocal cancelled
            try:
                async for _ in doubles:
                    assert False  # pragma: no cover
            except CancelledError:
                cancelled = True

        task = ensure_future(iterator_task())
        await sleep(0.05)
        assert not cancelled
        assert not doubles.is_closed
        assert iterator.value == 1
        assert not iterator.is_closed
        task.cancel()
        await sleep(0.05)
        assert cancelled
        assert iterator.value == -1
        assert doubles.is_closed
        assert iterator.is_closed



from typing import Any

from .location import SourceLocation

__all__ = ["Source", "is_source"]


class Source:
    """A representation of source input to GraphQL."""

    # allow custom attributes and weak references (not used internally)
    __slots__ = "__weakref__", "__dict__", "body", "name", "location_offset"

    def __init__(
        self,
        body: str,
        name: str = "GraphQL request",
        location_offset: SourceLocation = SourceLocation(1, 1),
    ) -> None:
        """Initialize source input.

        The ``name`` and ``location_offset`` parameters are optional, but they are
        useful for clients who store GraphQL documents in source files. For example,
        if the GraphQL input starts at line 40 in a file named ``Foo.graphql``, it might
        be useful for ``name`` to be ``"Foo.graphql"`` and location to be ``(40, 0)``.

        The ``line`` and ``column`` attributes in ``location_offset`` are 1-indexed.
        """
        self.body = body
        self.name = name
        if not isinstance(location_offset, SourceLocation):
            location_offset = SourceLocation._make(location_offset)
        if location_offset.line <= 0:
            raise ValueError(
                "line in location_offset is 1-indexed and must be positive."
            )
        if location_offset.column <= 0:
            raise ValueError(
                "column in location_offset is 1-indexed and must be positive."
            )
        self.location_offset = location_offset

    def get_location(self, position: int) -> SourceLocation:
        lines = self.body[:position].splitlines()
        if lines:
            line = len(lines)
            column = len(lines[-1]) + 1
        else:
            line = 1
            column = 1
        return SourceLocation(line, column)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r}>"

    def __eq__(self, other: Any) -> bool:
        return (isinstance(other, Source) and other.body == self.body) or (
            isinstance(other, str) and other == self.body
        )

    def __ne__(self, other: Any) -> bool:
        return not self == other


def is_source(source: Any) -> bool:
    """Test if the given value is a Source object.

    For internal use only.
    """
    return isinstance(source, Source)



from functools import partial

from graphql.validation import UniqueInputFieldNamesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, UniqueInputFieldNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_unique_input_field_names():
    def input_object_with_fields():
        assert_valid(
            """
            {
              field(arg: { f: true })
            }
            """
        )

    def same_input_object_within_two_args():
        assert_valid(
            """
            {
              field(arg1: { f: true }, arg2: { f: true })
            }
            """
        )

    def multiple_input_object_fields():
        assert_valid(
            """
            {
              field(arg: { f1: "value", f2: "value", f3: "value" })
            }
            """
        )

    def allows_for_nested_input_objects_with_similar_fields():
        assert_valid(
            """
            {
              field(arg: {
                deep: {
                  deep: {
                    id: 1
                  }
                  id: 1
                }
                id: 1
              })
            }
            """
        )

    def duplicate_input_object_fields():
        assert_errors(
            """
            {
              field(arg: { f1: "value", f1: "value" })
            }
            """,
            [
                {
                    "message": "There can be only one input field named 'f1'.",
                    "locations": [(3, 28), (3, 41)],
                },
            ],
        )

    def many_duplicate_input_object_fields():
        assert_errors(
            """
            {
              field(arg: { f1: "value", f1: "value", f1: "value" })
            }
            """,
            [
                {
                    "message": "There can be only one input field named 'f1'.",
                    "locations": [(3, 28), (3, 41)],
                },
                {
                    "message": "There can be only one input field named 'f1'.",
                    "locations": [(3, 28), (3, 54)],
                },
            ],
        )

    def nested_duplicate_input_object_fields():
        assert_errors(
            """
            {
              field(arg: { f1: {f2: "value", f2: "value" }})
            }
            """,
            [
                {
                    "message": "There can be only one input field named 'f2'.",
                    "locations": [(3, 33), (3, 46)],
                },
            ],
        )



from json import dumps
from typing import Optional

from pytest import raises

from graphql.error import GraphQLSyntaxError
from graphql.language import Lexer, Source, TokenKind, parse
from graphql.utilities import strip_ignored_characters

from ..fixtures import kitchen_sink_query, kitchen_sink_sdl  # noqa: F401
from ..utils import dedent

ignored_tokens = [
    # UnicodeBOM
    "\uFEFF",  # Byte Order Mark (U+FEFF)
    # WhiteSpace
    "\t",  # Horizontal Tab (U+0009)
    " ",  # Space (U+0020)
    # LineTerminator
    "\n",  # "New Line (U+000A)"
    "\r",  # "Carriage Return (U+000D)" [ lookahead ! "New Line (U+000A)" ]
    "\r\n",  # "Carriage Return (U+000D)" "New Line (U+000A)"
    # Comment
    '# "Comment" string\n',  # `#` CommentChar*
    # Comma
    ",",  # ,
]

punctuator_tokens = ["!", "$", "(", ")", "...", ":", "=", "@", "[", "]", "{", "|", "}"]

non_punctuator_tokens = [
    "name_token",  # Name
    "1",  # IntValue
    "3.14",  # FloatValue
    '"some string value"',  # StringValue
    '"""block\nstring\nvalue"""',  # StringValue(BlockString)
]


def lex_value(s: str) -> Optional[str]:
    lexer = Lexer(Source(s))
    value = lexer.advance().value
    assert lexer.advance().kind == TokenKind.EOF, "Expected EOF"
    return value


class ExpectStripped:
    def __init__(self, doc_string: str):
        self.doc_string = doc_string

    def to_equal(self, expected: str):
        doc_string = self.doc_string
        stripped = strip_ignored_characters(doc_string)

        assert stripped == expected, dedent(
            f"""
            Expected strip_ignored_characters({doc_string!r})
              to equal {expected!r}
              but got {stripped!r}
            """
        )

        stripped_twice = strip_ignored_characters(stripped)

        assert stripped == stripped_twice, dedent(
            f""""
            Expected strip_ignored_characters({stripped!r})"
              to equal {stripped!r}
              but got {stripped_twice!r}
            """
        )

    def to_stay_the_same(self):
        self.to_equal(self.doc_string)


def describe_strip_ignored_characters():
    def strips_ignored_characters_from_graphql_query_document():
        query = dedent(
            """
            query SomeQuery($foo: String!, $bar: String) {
              someField(foo: $foo, bar: $bar) {
                a
                b {
                  c
                  d
                }
              }
            }
            """
        )

        assert strip_ignored_characters(query) == (
            "query SomeQuery($foo:String!$bar:String)"
            "{someField(foo:$foo bar:$bar){a b{c d}}}"
        )

    def strips_ignored_characters_from_graphql_sdl_document():
        sdl = dedent(
            '''
            """
            Type description
            """
            type Foo {
              """
              Field description
              """
              bar: String
            }
          '''
        )

        assert strip_ignored_characters(sdl) == (
            '"""Type description""" type Foo{"""Field description""" bar:String}'
        )

    def report_document_with_invalid_token():
        with raises(GraphQLSyntaxError) as exc_info:
            strip_ignored_characters('{ foo(arg: "\n"')

        assert str(exc_info.value) == dedent(
            """
            Syntax Error: Unterminated string.

            GraphQL request:1:13
            1 | { foo(arg: "
              |             ^
            2 | "
            """
        )

    def strips_non_parsable_document():
        ExpectStripped('{ foo(arg: "str"').to_equal('{foo(arg:"str"')

    def strips_documents_with_only_ignored_characters():
        ExpectStripped("\n").to_equal("")
        ExpectStripped(",").to_equal("")
        ExpectStripped(",,").to_equal("")
        ExpectStripped("#comment\n, \n").to_equal("")

        for ignored in ignored_tokens:
            ExpectStripped(ignored).to_equal("")

            for another_ignored in ignored_tokens:
                ExpectStripped(ignored + another_ignored).to_equal("")

        ExpectStripped("".join(ignored_tokens)).to_equal("")

    def strips_leading_and_trailing_ignored_tokens():
        ExpectStripped("\n1").to_equal("1")
        ExpectStripped(",1").to_equal("1")
        ExpectStripped(",,1").to_equal("1")
        ExpectStripped("#comment\n, \n1").to_equal("1")

        ExpectStripped("1\n").to_equal("1")
        ExpectStripped("1,").to_equal("1")
        ExpectStripped("1,,").to_equal("1")
        ExpectStripped("1#comment\n, \n").to_equal("1")

        for token in punctuator_tokens + non_punctuator_tokens:
            for ignored in ignored_tokens:
                ExpectStripped(ignored + token).to_equal(token)
                ExpectStripped(token + ignored).to_equal(token)

                for another_ignored in ignored_tokens:
                    ExpectStripped(token + ignored + ignored).to_equal(token)
                    ExpectStripped(ignored + another_ignored + token).to_equal(token)

            ExpectStripped("".join(ignored_tokens) + token).to_equal(token)
            ExpectStripped(token + "".join(ignored_tokens)).to_equal(token)

    def strips_ignored_tokens_between_punctuator_tokens():
        ExpectStripped("[,)").to_equal("[)")
        ExpectStripped("[\r)").to_equal("[)")
        ExpectStripped("[\r\r)").to_equal("[)")
        ExpectStripped("[\r,)").to_equal("[)")
        ExpectStripped("[,\n)").to_equal("[)")

        for left in punctuator_tokens:
            for right in punctuator_tokens:
                for ignored in ignored_tokens:
                    ExpectStripped(left + ignored + right).to_equal(left + right)

                    for another_ignored in ignored_tokens:
                        ExpectStripped(
                            left + ignored + another_ignored + right
                        ).to_equal(left + right)

                ExpectStripped(left + "".join(ignored_tokens) + right).to_equal(
                    left + right
                )

    def strips_ignored_tokens_between_punctuator_and_non_punctuator_tokens():
        ExpectStripped("[,1").to_equal("[1")
        ExpectStripped("[\r1").to_equal("[1")
        ExpectStripped("[\r\r1").to_equal("[1")
        ExpectStripped("[\r,1").to_equal("[1")
        ExpectStripped("[,\n1").to_equal("[1")

        for non_punctuator in non_punctuator_tokens:
            for punctuator in punctuator_tokens:
                for ignored in ignored_tokens:
                    ExpectStripped(punctuator + ignored + non_punctuator).to_equal(
                        punctuator + non_punctuator
                    )

                    for another_ignored in ignored_tokens:
                        ExpectStripped(
                            punctuator + ignored + another_ignored + non_punctuator
                        ).to_equal(punctuator + non_punctuator)

                ExpectStripped(
                    punctuator + "".join(ignored_tokens) + non_punctuator
                ).to_equal(punctuator + non_punctuator)

    def strips_ignored_tokens_between_non_punctuator_and_punctuator_tokens():
        ExpectStripped("1,[").to_equal("1[")
        ExpectStripped("1\r[").to_equal("1[")
        ExpectStripped("1\r\r[").to_equal("1[")
        ExpectStripped("1\r,[").to_equal("1[")
        ExpectStripped("1,\n[").to_equal("1[")

        for non_punctuator in non_punctuator_tokens:
            for punctuator in punctuator_tokens:
                # Special case for that is handled in the below test
                if punctuator == "...":
                    continue

                for ignored in ignored_tokens:
                    ExpectStripped(non_punctuator + ignored + punctuator).to_equal(
                        non_punctuator + punctuator
                    )

                    for another_ignored in ignored_tokens:
                        ExpectStripped(
                            non_punctuator + ignored + another_ignored + punctuator
                        ).to_equal(non_punctuator + punctuator)

                ExpectStripped(
                    non_punctuator + "".join(ignored_tokens) + punctuator
                ).to_equal(non_punctuator + punctuator)

    def replace_ignored_tokens_between_non_punctuator_tokens_and_spread_with_space():
        ExpectStripped("a ...").to_equal("a ...")
        ExpectStripped("1 ...").to_equal("1 ...")
        ExpectStripped("1 ... ...").to_equal("1 ......")

        for non_punctuator in non_punctuator_tokens:
            for ignored in ignored_tokens:
                ExpectStripped(non_punctuator + ignored + "...").to_equal(
                    non_punctuator + " ..."
                )

                for another_ignored in ignored_tokens:
                    ExpectStripped(
                        non_punctuator + ignored + another_ignored + " ..."
                    ).to_equal(non_punctuator + " ...")

            ExpectStripped(non_punctuator + "".join(ignored_tokens) + "...").to_equal(
                non_punctuator + " ..."
            )

    def replace_ignored_tokens_between_non_punctuator_tokens_with_space():
        ExpectStripped("1 2").to_stay_the_same()
        ExpectStripped('"" ""').to_stay_the_same()
        ExpectStripped("a b").to_stay_the_same()

        ExpectStripped("a,1").to_equal("a 1")
        ExpectStripped("a,,1").to_equal("a 1")
        ExpectStripped("a  1").to_equal("a 1")
        ExpectStripped("a \t 1").to_equal("a 1")

        for left in non_punctuator_tokens:
            for right in non_punctuator_tokens:
                for ignored in ignored_tokens:
                    ExpectStripped(left + ignored + right).to_equal(left + " " + right)

                    for another_ignored in ignored_tokens:
                        ExpectStripped(
                            left + ignored + another_ignored + right
                        ).to_equal(left + " " + right)

                ExpectStripped(left + "".join(ignored_tokens) + right).to_equal(
                    left + " " + right
                )

    def does_not_strip_ignored_tokens_embedded_in_the_string():
        ExpectStripped('" "').to_stay_the_same()
        ExpectStripped('","').to_stay_the_same()
        ExpectStripped('",,"').to_stay_the_same()
        ExpectStripped('",|"').to_stay_the_same()

        for ignored in ignored_tokens:
            ExpectStripped(dumps(ignored)).to_stay_the_same()

            for another_ignored in ignored_tokens:
                ExpectStripped(dumps(ignored + another_ignored)).to_stay_the_same()

        ExpectStripped(dumps("".join(ignored_tokens))).to_stay_the_same()

    def does_not_strip_ignored_tokens_embedded_in_the_block_string():
        ExpectStripped('""","""').to_stay_the_same()
        ExpectStripped('""",,"""').to_stay_the_same()
        ExpectStripped('""",|"""').to_stay_the_same()

        ignored_tokens_without_formatting = [
            token
            for token in ignored_tokens
            if token not in ["\n", "\r", "\r\n", "\t", " "]
        ]

        for ignored in ignored_tokens_without_formatting:
            ExpectStripped('"""|' + ignored + '|"""').to_stay_the_same()

            for another_ignored in ignored_tokens_without_formatting:
                ExpectStripped(
                    '"""|' + ignored + another_ignored + '|"""'
                ).to_stay_the_same()

        ExpectStripped(
            '"""|' + "".join(ignored_tokens_without_formatting) + '|"""'
        ).to_stay_the_same()

    def strips_ignored_characters_inside_block_strings():
        # noinspection PyShadowingNames
        def expect_stripped_string(block_str: str):
            original_value = lex_value(block_str)
            stripped_value = lex_value(strip_ignored_characters(block_str))

            assert original_value == stripped_value, dedent(
                f"""
                Expected lexValue(stripIgnoredCharacters({block_str!r})
                  to equal {original_value!r}
                  but got {stripped_value!r}
                """
            )
            return ExpectStripped(block_str)

        expect_stripped_string('""""""').to_stay_the_same()
        expect_stripped_string('""" """').to_equal('""""""')

        expect_stripped_string('"""a"""').to_stay_the_same()
        expect_stripped_string('""" a"""').to_equal('""" a"""')
        expect_stripped_string('""" a """').to_equal('""" a """')

        expect_stripped_string('"""\n"""').to_equal('""""""')
        expect_stripped_string('"""a\nb"""').to_equal('"""a\nb"""')
        expect_stripped_string('"""a\rb"""').to_equal('"""a\nb"""')
        expect_stripped_string('"""a\r\nb"""').to_equal('"""a\nb"""')
        expect_stripped_string('"""a\r\n\nb"""').to_equal('"""a\n\nb"""')

        expect_stripped_string('"""\\\n"""').to_stay_the_same()
        expect_stripped_string('""""\n"""').to_stay_the_same()
        expect_stripped_string('"""\\"""\n"""').to_equal('"""\\""""""')

        expect_stripped_string('"""\na\n b"""').to_stay_the_same()
        expect_stripped_string('"""\n a\n b"""').to_equal('"""a\nb"""')
        expect_stripped_string('"""\na\n b\nc"""').to_equal('"""a\n b\nc"""')

    # noinspection PyShadowingNames
    def strips_kitchen_sink_query_but_maintains_the_exact_same_ast(
        kitchen_sink_query,  # noqa: F811
    ):
        stripped_query = strip_ignored_characters(kitchen_sink_query)
        assert strip_ignored_characters(stripped_query) == stripped_query

        query_ast = parse(kitchen_sink_query, no_location=True)
        stripped_ast = parse(stripped_query, no_location=True)
        assert stripped_ast == query_ast

    # noinspection PyShadowingNames
    def strips_kitchen_sink_sdl_but_maintains_the_exact_same_ast(
        kitchen_sink_sdl,  # noqa: F811
    ):
        stripped_sdl = strip_ignored_characters(kitchen_sink_sdl)
        assert strip_ignored_characters(stripped_sdl) == stripped_sdl

        sdl_ast = parse(kitchen_sink_sdl, no_location=True)
        stripped_ast = parse(stripped_sdl, no_location=True)
        assert stripped_ast == sdl_ast



from graphql.language import parse, print_ast, Source
from graphql.utilities import concat_ast

from ..utils import dedent


def describe_concat_ast():
    def concatenates_two_asts_together():
        source_a = Source(
            """
            { a, b, ... Frag }
            """
        )

        source_b = Source(
            """
            fragment Frag on T {
                c
            }
            """
        )

        ast_a = parse(source_a)
        ast_b = parse(source_b)
        ast_c = concat_ast([ast_a, ast_b])

        assert print_ast(ast_c) == dedent(
            """
            {
              a
              b
              ...Frag
            }

            fragment Frag on T {
              c
            }
            """
        )



from typing import Optional

from ..language import DocumentNode, OperationDefinitionNode

__all__ = ["get_operation_ast"]


def get_operation_ast(
    document_ast: DocumentNode, operation_name: Optional[str] = None
) -> Optional[OperationDefinitionNode]:
    """Get operation AST node.

    Returns an operation AST given a document AST and optionally an operation
    name. If a name is not provided, an operation is only returned if only one
    is provided in the document.
    """
    operation = None
    for definition in document_ast.definitions:
        if isinstance(definition, OperationDefinitionNode):
            if operation_name is None:
                # If no operation name was provided, only return an Operation if there
                # is one defined in the document.
                # Upon encountering the second, return None.
                if operation:
                    return None
                operation = definition
            elif definition.name and definition.name.value == operation_name:
                return definition
    return operation



"""Python Utils

This package contains dependency-free Python utility functions used throughout the
codebase.

Each utility should belong in its own file and be the default export.

These functions are not part of the module interface and are subject to change.
"""

from .convert_case import camel_to_snake, snake_to_camel
from .cached_property import cached_property
from .description import (
    Description,
    is_description,
    register_description,
    unregister_description,
)
from .did_you_mean import did_you_mean
from .group_by import group_by
from .identity_func import identity_func
from .inspect import inspect
from .is_awaitable import is_awaitable
from .is_iterable import is_collection, is_iterable
from .natural_compare import natural_comparison_key
from .awaitable_or_value import AwaitableOrValue
from .suggestion_list import suggestion_list
from .frozen_error import FrozenError
from .frozen_list import FrozenList
from .frozen_dict import FrozenDict
from .merge_kwargs import merge_kwargs
from .path import Path
from .print_path_list import print_path_list
from .simple_pub_sub import SimplePubSub, SimplePubSubIterator
from .undefined import Undefined, UndefinedType

__all__ = [
    "camel_to_snake",
    "snake_to_camel",
    "cached_property",
    "did_you_mean",
    "Description",
    "group_by",
    "is_description",
    "register_description",
    "unregister_description",
    "identity_func",
    "inspect",
    "is_awaitable",
    "is_collection",
    "is_iterable",
    "merge_kwargs",
    "natural_comparison_key",
    "AwaitableOrValue",
    "suggestion_list",
    "FrozenError",
    "FrozenList",
    "FrozenDict",
    "Path",
    "print_path_list",
    "SimplePubSub",
    "SimplePubSubIterator",
    "Undefined",
    "UndefinedType",
]



from math import inf, nan, pi
from typing import Any

from pytest import raises

from graphql.error import GraphQLError
from graphql.language import parse_value as parse_value_to_ast
from graphql.pyutils import Undefined
from graphql.type import (
    GraphQLInt,
    GraphQLFloat,
    GraphQLString,
    GraphQLBoolean,
    GraphQLID,
)


def describe_type_system_specified_scalar_types():
    def describe_graphql_int():
        def parse_value():
            _parse_value = GraphQLInt.parse_value

            def _parse_value_raises(s: Any, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_value(s)
                assert str(exc_info.value) == message

            assert _parse_value(1) == 1
            assert _parse_value(0) == 0
            assert _parse_value(-1) == -1

            _parse_value_raises(
                9876504321,
                "Int cannot represent non 32-bit signed integer value: 9876504321",
            )
            _parse_value_raises(
                -9876504321,
                "Int cannot represent non 32-bit signed integer value: -9876504321",
            )
            _parse_value_raises(0.1, "Int cannot represent non-integer value: 0.1")
            _parse_value_raises(nan, "Int cannot represent non-integer value: nan")
            _parse_value_raises(inf, "Int cannot represent non-integer value: inf")
            _parse_value_raises(
                Undefined, "Int cannot represent non-integer value: Undefined"
            )
            _parse_value_raises(None, "Int cannot represent non-integer value: None")
            _parse_value_raises("", "Int cannot represent non-integer value: ''")
            _parse_value_raises("123", "Int cannot represent non-integer value: '123'")
            _parse_value_raises(False, "Int cannot represent non-integer value: False")
            _parse_value_raises(True, "Int cannot represent non-integer value: True")
            _parse_value_raises([1], "Int cannot represent non-integer value: [1]")
            _parse_value_raises(
                {"value": 1}, "Int cannot represent non-integer value: {'value': 1}"
            )

        def parse_literal():
            def _parse_literal(s: str):
                return GraphQLInt.parse_literal(parse_value_to_ast(s))

            def _parse_literal_raises(s: str, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_literal(s)
                assert str(exc_info.value).startswith(message + "\n")

            assert _parse_literal("1") == 1
            assert _parse_literal("0") == 0
            assert _parse_literal("-1") == -1

            _parse_literal_raises(
                "9876504321",
                "Int cannot represent non 32-bit signed integer value: 9876504321",
            )
            _parse_literal_raises(
                "-9876504321",
                "Int cannot represent non 32-bit signed integer value: -9876504321",
            )
            _parse_literal_raises("1.0", "Int cannot represent non-integer value: 1.0")
            _parse_literal_raises(
                "null", "Int cannot represent non-integer value: null"
            )
            _parse_literal_raises(
                "None", "Int cannot represent non-integer value: None"
            )
            _parse_literal_raises('""', 'Int cannot represent non-integer value: ""')
            _parse_literal_raises(
                '"123"', 'Int cannot represent non-integer value: "123"'
            )
            _parse_literal_raises(
                "false", "Int cannot represent non-integer value: false"
            )
            _parse_literal_raises(
                "False", "Int cannot represent non-integer value: False"
            )
            _parse_literal_raises("[1]", "Int cannot represent non-integer value: [1]")
            _parse_literal_raises(
                "{value: 1}", "Int cannot represent non-integer value: {value: 1}"
            )
            _parse_literal_raises(
                "ENUM_VALUE", "Int cannot represent non-integer value: ENUM_VALUE"
            )
            _parse_literal_raises(
                "$var", "Int cannot represent non-integer value: $var"
            )

        def serializes():
            serialize = GraphQLInt.serialize

            assert serialize(1) == 1
            assert serialize("123") == 123
            assert serialize(0) == 0
            assert serialize(-1) == -1
            assert serialize(1e5) == 100000
            assert serialize(False) == 0
            assert serialize(True) == 1
            assert serialize(type("Int", (int,), {})(5)) == 5

            # The GraphQL specification does not allow serializing non-integer
            # values as Int to avoid accidental data loss.
            with raises(GraphQLError) as exc_info:
                serialize(0.1)
            assert str(exc_info.value) == "Int cannot represent non-integer value: 0.1"
            with raises(GraphQLError) as exc_info:
                serialize(1.1)
            assert str(exc_info.value) == "Int cannot represent non-integer value: 1.1"
            with raises(GraphQLError) as exc_info:
                serialize(-1.1)
            assert str(exc_info.value) == "Int cannot represent non-integer value: -1.1"
            with raises(GraphQLError) as exc_info:
                serialize("-1.1")
            assert (
                str(exc_info.value) == "Int cannot represent non-integer value: '-1.1'"
            )
            # Maybe a safe JavaScript int, but bigger than 2^32, so not
            # representable as a GraphQL Int
            with raises(GraphQLError) as exc_info:
                serialize(9876504321)
            assert str(exc_info.value) == (
                "Int cannot represent non 32-bit signed integer value: 9876504321"
            )
            with raises(GraphQLError) as exc_info:
                serialize(-9876504321)
            assert str(exc_info.value) == (
                "Int cannot represent non 32-bit signed integer value: -9876504321"
            )
            # Too big to represent as an Int in JavaScript or GraphQL
            with raises(GraphQLError) as exc_info:
                serialize(1e100)
            assert str(exc_info.value) == (
                "Int cannot represent non 32-bit signed integer value: 1e+100"
            )
            with raises(GraphQLError) as exc_info:
                serialize(-1e100)
            assert str(exc_info.value) == (
                "Int cannot represent non 32-bit signed integer value: -1e+100"
            )
            with raises(GraphQLError) as exc_info:
                serialize("one")
            assert (
                str(exc_info.value) == "Int cannot represent non-integer value: 'one'"
            )
            # Doesn't represent number
            with raises(GraphQLError) as exc_info:
                serialize("")
            assert str(exc_info.value) == "Int cannot represent non-integer value: ''"
            with raises(GraphQLError) as exc_info:
                serialize(nan)
            assert str(exc_info.value) == "Int cannot represent non-integer value: nan"
            with raises(GraphQLError) as exc_info:
                serialize(inf)
            assert str(exc_info.value) == "Int cannot represent non-integer value: inf"
            with raises(GraphQLError) as exc_info:
                serialize([5])
            assert str(exc_info.value) == "Int cannot represent non-integer value: [5]"

    def describe_graphql_float():
        def parse_value():
            _parse_value = GraphQLFloat.parse_value

            def _parse_value_raises(s: Any, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_value(s)
                assert str(exc_info.value) == message

            assert _parse_value(1) == 1
            assert _parse_value(0) == 0
            assert _parse_value(-1) == -1
            assert _parse_value(0.1) == 0.1
            assert _parse_value(pi) == pi

            _parse_value_raises(nan, "Float cannot represent non numeric value: nan")
            _parse_value_raises(inf, "Float cannot represent non numeric value: inf")
            _parse_value_raises("", "Float cannot represent non numeric value: ''")
            _parse_value_raises(
                "123", "Float cannot represent non numeric value: '123'"
            )
            _parse_value_raises(
                "123.5", "Float cannot represent non numeric value: '123.5'"
            )
            _parse_value_raises(
                False, "Float cannot represent non numeric value: False"
            )
            _parse_value_raises(True, "Float cannot represent non numeric value: True")
            _parse_value_raises(
                [0.1], "Float cannot represent non numeric value: [0.1]"
            )
            _parse_value_raises(
                {"value": 0.1},
                "Float cannot represent non numeric value: {'value': 0.1}",
            )

        def parse_literal():
            def _parse_literal(s: str):
                return GraphQLFloat.parse_literal(parse_value_to_ast(s))

            def _parse_literal_raises(s: str, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_literal(s)
                assert str(exc_info.value).startswith(message + "\n")

            assert _parse_literal("1") == 1
            assert _parse_literal("0") == 0
            assert _parse_literal("-1") == -1
            assert _parse_literal("0.1") == 0.1
            assert _parse_literal(str(pi)) == pi

            _parse_literal_raises(
                "null", "Float cannot represent non numeric value: null"
            )
            _parse_literal_raises(
                "None", "Float cannot represent non numeric value: None"
            )
            _parse_literal_raises('""', 'Float cannot represent non numeric value: ""')
            _parse_literal_raises(
                '"123"', 'Float cannot represent non numeric value: "123"'
            )
            _parse_literal_raises(
                '"123.5"', 'Float cannot represent non numeric value: "123.5"'
            )
            _parse_literal_raises(
                "false", "Float cannot represent non numeric value: false"
            )
            _parse_literal_raises(
                "False", "Float cannot represent non numeric value: False"
            )
            _parse_literal_raises(
                "[0.1]", "Float cannot represent non numeric value: [0.1]"
            )
            _parse_literal_raises(
                "{value: 0.1}", "Float cannot represent non numeric value: {value: 0.1}"
            )
            _parse_literal_raises(
                "ENUM_VALUE", "Float cannot represent non numeric value: ENUM_VALUE"
            )
            _parse_literal_raises(
                "$var", "Float cannot represent non numeric value: $var"
            )

        def serializes():
            serialize = GraphQLFloat.serialize

            assert serialize(1) == 1.0
            assert serialize(0) == 0.0
            assert serialize("123.5") == 123.5
            assert serialize(-1) == -1.0
            assert serialize(0.1) == 0.1
            assert serialize(1.1) == 1.1
            assert serialize(-1.1) == -1.1
            assert serialize("-1.1") == -1.1
            assert serialize(False) == 0
            assert serialize(True) == 1
            assert serialize(type("Float", (float,), {})(5.5)) == 5.5

            with raises(GraphQLError) as exc_info:
                serialize(nan)
            assert (
                str(exc_info.value) == "Float cannot represent non numeric value: nan"
            )
            with raises(GraphQLError) as exc_info:
                serialize(inf)
            assert (
                str(exc_info.value) == "Float cannot represent non numeric value: inf"
            )
            with raises(GraphQLError) as exc_info:
                serialize("one")
            assert str(exc_info.value) == (
                "Float cannot represent non numeric value: 'one'"
            )
            with raises(GraphQLError) as exc_info:
                serialize("")
            assert str(exc_info.value) == "Float cannot represent non numeric value: ''"
            with raises(GraphQLError) as exc_info:
                serialize([5])
            assert (
                str(exc_info.value) == "Float cannot represent non numeric value: [5]"
            )

    def describe_graphql_string():
        def parse_value():
            _parse_value = GraphQLString.parse_value

            def _parse_value_raises(s: Any, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_value(s)
                assert str(exc_info.value) == message

            assert _parse_value("foo") == "foo"

            _parse_value_raises(
                Undefined, "String cannot represent a non string value: Undefined"
            )
            _parse_value_raises(
                None, "String cannot represent a non string value: None"
            )
            _parse_value_raises(1, "String cannot represent a non string value: 1")
            _parse_value_raises(nan, "String cannot represent a non string value: nan")
            _parse_value_raises(
                False, "String cannot represent a non string value: False"
            )
            _parse_value_raises(
                ["foo"], "String cannot represent a non string value: ['foo']"
            )
            _parse_value_raises(
                {"value": "foo"},
                "String cannot represent a non string value: {'value': 'foo'}",
            )

        def parse_literal():
            def _parse_literal(s: str):
                return GraphQLString.parse_literal(parse_value_to_ast(s))

            def _parse_literal_raises(s: str, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_literal(s)
                assert str(exc_info.value).startswith(message + "\n")

            assert _parse_literal('"foo"') == "foo"
            assert _parse_literal('"""bar"""') == "bar"

            _parse_literal_raises(
                "null", "String cannot represent a non string value: null"
            )
            _parse_literal_raises(
                "None", "String cannot represent a non string value: None"
            )
            _parse_literal_raises("1", "String cannot represent a non string value: 1")
            _parse_literal_raises(
                "0.1", "String cannot represent a non string value: 0.1"
            )
            _parse_literal_raises(
                "false", "String cannot represent a non string value: false"
            )
            _parse_literal_raises(
                "False", "String cannot represent a non string value: False"
            )
            _parse_literal_raises(
                '["foo"]', 'String cannot represent a non string value: ["foo"]'
            )
            _parse_literal_raises(
                '{value: "foo"}',
                'String cannot represent a non string value: {value: "foo"}',
            )
            _parse_literal_raises(
                "ENUM_VALUE", "String cannot represent a non string value: ENUM_VALUE"
            )
            _parse_literal_raises(
                "$var", "String cannot represent a non string value: $var"
            )

        def serializes():
            serialize = GraphQLString.serialize

            assert serialize("string") == "string"
            assert serialize(1) == "1"
            assert serialize(-1.1) == "-1.1"
            assert serialize(True) == "true"
            assert serialize(False) == "false"

            class StringableObjValue:
                def __str__(self):
                    return "something useful"

            assert serialize(StringableObjValue()) == "something useful"

            with raises(GraphQLError) as exc_info:
                serialize(nan)
            assert str(exc_info.value) == "String cannot represent value: nan"

            with raises(GraphQLError) as exc_info:
                serialize([1])
            assert str(exc_info.value) == "String cannot represent value: [1]"

            with raises(GraphQLError) as exc_info:
                serialize({})
            assert str(exc_info.value) == "String cannot represent value: {}"

            with raises(GraphQLError) as exc_info:
                serialize({"value_of": "value_of string"})
            assert (
                str(exc_info.value) == "String cannot represent value:"
                " {'value_of': 'value_of string'}"
            )

    def describe_graphql_boolean():
        def parse_value():
            _parse_value = GraphQLBoolean.parse_value

            def _parse_value_raises(s: Any, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_value(s)
                assert str(exc_info.value) == message

            assert _parse_value(True) is True
            assert _parse_value(False) is False

            _parse_value_raises(
                Undefined, "Boolean cannot represent a non boolean value: Undefined"
            )
            _parse_value_raises(
                None, "Boolean cannot represent a non boolean value: None"
            )
            _parse_value_raises(0, "Boolean cannot represent a non boolean value: 0")
            _parse_value_raises(1, "Boolean cannot represent a non boolean value: 1")
            _parse_value_raises(
                nan, "Boolean cannot represent a non boolean value: nan"
            )
            _parse_value_raises("", "Boolean cannot represent a non boolean value: ''")
            _parse_value_raises(
                "false", "Boolean cannot represent a non boolean value: 'false'"
            )
            _parse_value_raises(
                "False", "Boolean cannot represent a non boolean value: 'False'"
            )
            _parse_value_raises(
                [False], "Boolean cannot represent a non boolean value: [False]"
            )
            _parse_value_raises(
                {"value": False},
                "Boolean cannot represent a non boolean value: {'value': False}",
            )

        def parse_literal():
            def _parse_literal(s: str):
                return GraphQLBoolean.parse_literal(parse_value_to_ast(s))

            def _parse_literal_raises(s: str, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_literal(s)
                assert str(exc_info.value).startswith(message + "\n")

            assert _parse_literal("true") is True
            assert _parse_literal("false") is False

            _parse_literal_raises(
                "True", "Boolean cannot represent a non boolean value: True"
            )
            _parse_literal_raises(
                "False", "Boolean cannot represent a non boolean value: False"
            )
            _parse_literal_raises(
                "null", "Boolean cannot represent a non boolean value: null"
            )
            _parse_literal_raises(
                "None", "Boolean cannot represent a non boolean value: None"
            )
            _parse_literal_raises(
                "0", "Boolean cannot represent a non boolean value: 0"
            )
            _parse_literal_raises(
                "1", "Boolean cannot represent a non boolean value: 1"
            )
            _parse_literal_raises(
                "0.1", "Boolean cannot represent a non boolean value: 0.1"
            )
            _parse_literal_raises(
                '""', 'Boolean cannot represent a non boolean value: ""'
            )
            _parse_literal_raises(
                '"false"', 'Boolean cannot represent a non boolean value: "false"'
            )
            _parse_literal_raises(
                '"False"', 'Boolean cannot represent a non boolean value: "False"'
            )
            _parse_literal_raises(
                "[false]", "Boolean cannot represent a non boolean value: [false]"
            )
            _parse_literal_raises(
                "[False]", "Boolean cannot represent a non boolean value: [False]"
            )
            _parse_literal_raises(
                "{value: false}",
                "Boolean cannot represent a non boolean value: {value: false}",
            )
            _parse_literal_raises(
                "{value: False}",
                "Boolean cannot represent a non boolean value: {value: False}",
            )
            _parse_literal_raises(
                "ENUM_VALUE", "Boolean cannot represent a non boolean value: ENUM_VALUE"
            )
            _parse_literal_raises(
                "$var", "Boolean cannot represent a non boolean value: $var"
            )

        def serializes():
            serialize = GraphQLBoolean.serialize

            assert serialize(1) is True
            assert serialize(0) is False
            assert serialize(True) is True
            assert serialize(False) is False
            with raises(TypeError, match="not an acceptable base type"):
                # you can't subclass bool in Python
                assert serialize(type("Boolean", (bool,), {})(True)) is True

            with raises(GraphQLError) as exc_info:
                serialize(nan)
            assert str(exc_info.value) == (
                "Boolean cannot represent a non boolean value: nan"
            )

            with raises(GraphQLError) as exc_info:
                serialize("")
            assert str(exc_info.value) == (
                "Boolean cannot represent a non boolean value: ''"
            )

            with raises(GraphQLError) as exc_info:
                serialize("True")
            assert str(exc_info.value) == (
                "Boolean cannot represent a non boolean value: 'True'"
            )

            with raises(GraphQLError) as exc_info:
                serialize([False])
            assert str(exc_info.value) == (
                "Boolean cannot represent a non boolean value: [False]"
            )

            with raises(GraphQLError) as exc_info:
                serialize({})
            assert str(exc_info.value) == (
                "Boolean cannot represent a non boolean value: {}"
            )

    def describe_graphql_id():
        def parse_value():
            _parse_value = GraphQLID.parse_value

            def _parse_value_raises(s: Any, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_value(s)
                assert str(exc_info.value) == message

            assert _parse_value("") == ""
            assert _parse_value("1") == "1"
            assert _parse_value("foo") == "foo"
            assert _parse_value(1) == "1"
            assert _parse_value(0) == "0"
            assert _parse_value(-1) == "-1"
            assert _parse_value(1.0) == "1"

            # Maximum and minimum safe numbers in JS
            assert _parse_value(9007199254740991) == "9007199254740991"
            assert _parse_value(-9007199254740991) == "-9007199254740991"

            _parse_value_raises(Undefined, "ID cannot represent value: Undefined")
            _parse_value_raises(None, "ID cannot represent value: None")
            _parse_value_raises(0.1, "ID cannot represent value: 0.1")
            _parse_value_raises(nan, "ID cannot represent value: nan")
            _parse_value_raises(inf, "ID cannot represent value: inf")
            _parse_value_raises(False, "ID cannot represent value: False")
            _parse_value_raises(["1"], "ID cannot represent value: ['1']")
            _parse_value_raises(
                {"value": "1"}, "ID cannot represent value: {'value': '1'}"
            )

        def parse_literal():
            def _parse_literal(s: str):
                return GraphQLID.parse_literal(parse_value_to_ast(s))

            def _parse_literal_raises(s: str, message: str):
                with raises(GraphQLError) as exc_info:
                    _parse_literal(s)
                assert str(exc_info.value).startswith(message + "\n")

            assert _parse_literal('""') == ""
            assert _parse_literal('"1"') == "1"
            assert _parse_literal('"foo"') == "foo"
            assert _parse_literal('"""foo"""') == "foo"
            assert _parse_literal("1") == "1"
            assert _parse_literal("0") == "0"
            assert _parse_literal("-1") == "-1"

            # Support arbitrary long numbers even if they can't be represented in JS
            assert _parse_literal("90071992547409910") == "90071992547409910"
            assert _parse_literal("-90071992547409910") == "-90071992547409910"

            _parse_literal_raises(
                "null", "ID cannot represent a non-string and non-integer value: null"
            )
            _parse_literal_raises(
                "None", "ID cannot represent a non-string and non-integer value: None"
            )
            _parse_literal_raises(
                "0.1", "ID cannot represent a non-string and non-integer value: 0.1"
            )
            _parse_literal_raises(
                "false", "ID cannot represent a non-string and non-integer value: false"
            )
            _parse_literal_raises(
                "False", "ID cannot represent a non-string and non-integer value: False"
            )
            _parse_literal_raises(
                '["1"]', 'ID cannot represent a non-string and non-integer value: ["1"]'
            )
            _parse_literal_raises(
                '{value: "1"}',
                "ID cannot represent a non-string and non-integer value:"
                ' {value: "1"}',
            )
            _parse_literal_raises(
                "ENUM_VALUE",
                "ID cannot represent a non-string and non-integer value: ENUM_VALUE",
            )
            _parse_literal_raises(
                "$var", "ID cannot represent a non-string and non-integer value: $var"
            )

        def serializes():
            serialize = GraphQLID.serialize

            assert serialize("string") == "string"
            assert serialize("false") == "false"
            assert serialize("") == ""
            assert serialize(123) == "123"
            assert serialize(42.0) == "42"
            assert serialize(0) == "0"
            assert serialize(-1) == "-1"

            class ObjValue:
                def __init__(self, value):
                    self._id = value

                def __str__(self):
                    return str(self._id)

            obj_value = ObjValue(123)
            assert serialize(obj_value) == "123"

            with raises(GraphQLError) as exc_info:
                serialize(True)
            assert str(exc_info.value) == "ID cannot represent value: True"

            with raises(GraphQLError) as exc_info:
                serialize(3.14)
            assert str(exc_info.value) == "ID cannot represent value: 3.14"

            with raises(GraphQLError) as exc_info:
                serialize({})
            assert str(exc_info.value) == "ID cannot represent value: {}"

            with raises(GraphQLError) as exc_info:
                serialize(["abc"])
            assert str(exc_info.value) == "ID cannot represent value: ['abc']"



from functools import partial

from graphql.validation import LoneAnonymousOperationRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, LoneAnonymousOperationRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_anonymous_operation_must_be_alone():
    def no_operations():
        assert_valid(
            """
            fragment fragA on Type {
              field
            }
            """
        )

    def one_anon_operation():
        assert_valid(
            """
            {
              field
            }
            """
        )

    def multiple_named_operation():
        assert_valid(
            """
            query Foo {
              field
            }

            query Bar {
              field
            }
            """
        )

    def anon_operation_with_fragment():
        assert_valid(
            """
            {
              ...Foo
            }
            fragment Foo on Type {
              field
            }
            """
        )

    def multiple_anon_operations():
        assert_errors(
            """
            {
              fieldA
            }
            {
              fieldB
            }
            """,
            [
                {
                    "message": "This anonymous operation"
                    " must be the only defined operation.",
                    "locations": [(2, 13)],
                },
                {
                    "message": "This anonymous operation"
                    " must be the only defined operation.",
                    "locations": [(5, 13)],
                },
            ],
        )

    def anon_operation_with_a_mutation():
        assert_errors(
            """
            {
              fieldA
            }
            mutation Foo {
              fieldB
            }
            """,
            [
                {
                    "message": "This anonymous operation"
                    " must be the only defined operation.",
                    "locations": [(2, 13)],
                },
            ],
        )

    def anon_operation_with_a_subscription():
        assert_errors(
            """
            {
              fieldA
            }
            subscription Foo {
              fieldB
            }
            """,
            [
                {
                    "message": "This anonymous operation"
                    " must be the only defined operation.",
                    "locations": [(2, 13)],
                },
            ],
        )



from . import dedent


def describe_dedent():
    def removes_indentation_in_typical_usage():
        assert (
            dedent(
                """
                type Query {
                  me: User
                }

                type User {
                  id: ID
                  name: String
                }
                """
            )
            == "type Query {\n  me: User\n}\n\n"
            "type User {\n  id: ID\n  name: String\n}"
        )

    def removes_only_the_first_level_of_indentation():
        assert (
            dedent(
                """
                first
                  second
                    third
                      fourth
                """
            )
            == "first\n  second\n    third\n      fourth"
        )

    def does_not_escape_special_characters():
        assert (
            dedent(
                """
                type Root {
                  field(arg: String = "wi\th de\fault"): String
                }
                """
            )
            == "type Root {\n"
            '  field(arg: String = "wi\th de\fault"): String\n}'
        )

    def also_removes_indentation_using_tabs():
        assert (
            dedent(
                """
                \t\t    type Query {
                \t\t      me: User
                \t\t    }
                """
            )
            == "type Query {\n  me: User\n}"
        )

    def removes_leading_and_trailing_newlines():
        assert (
            dedent(
                """


                 type Query {
                   me: User
                 }


                 """
            )
            == "type Query {\n  me: User\n}"
        )

    def removes_all_trailing_spaces_and_tabs():
        assert (
            dedent(
                """
                type Query {
                  me: User
                }
                    \t\t  \t """
            )
            == "type Query {\n  me: User\n}"
        )

    def works_on_text_without_leading_newline():
        assert (
            dedent(
                """                type Query {
                  me: User
                }
                """
            )
            == "type Query {\n  me: User\n}"
        )



from typing import cast, Any, Awaitable

from pytest import mark

from graphql.execution import execute, execute_sync, ExecutionResult
from graphql.language import parse
from graphql.pyutils import is_awaitable
from graphql.utilities import build_schema


class Data:
    def __init__(self, value):
        self.listField = value


async def get_async(value):
    return value


def describe_execute_accepts_any_iterable_as_list_value():
    def _complete(list_field):
        return execute_sync(
            build_schema("type Query { listField: [String] }"),
            parse("{ listField }"),
            Data(list_field),
        )

    def accepts_a_set_as_a_list_value():
        # Note that sets are not ordered in Python.
        list_field = {"apple", "banana", "coconut"}
        result = _complete(list_field)
        assert result.errors is None
        assert isinstance(result.data, dict)
        assert list(result.data) == ["listField"]
        assert isinstance(result.data["listField"], list)
        assert set(result.data["listField"]) == list_field

    def accepts_a_generator_as_a_list_value():
        def list_field():
            yield "one"
            yield 2
            yield True

        assert _complete(list_field()) == (
            {"listField": ["one", "2", "true"]},
            None,
        )

    def accepts_a_custom_iterable_as_a_list_value():
        class ListField:
            def __iter__(self):
                self.last = "hello"
                return self

            def __next__(self):
                last = self.last
                if last == "stop":
                    raise StopIteration
                self.last = "world" if last == "hello" else "stop"
                return last

        assert _complete(ListField()) == (
            {"listField": ["hello", "world"]},
            None,
        )

    def accepts_function_arguments_as_a_list_value():
        def get_args(*args):
            return args  # actually just a tuple, nothing special in Python

        assert _complete(get_args("one", "two")) == (
            {"listField": ["one", "two"]},
            None,
        )

    def does_not_accept_a_dict_as_a_list_value():
        assert _complete({1: "one", 2: "two"}) == (
            {"listField": None},
            [
                {
                    "message": "Expected Iterable,"
                    " but did not find one for field 'Query.listField'.",
                    "locations": [(1, 3)],
                    "path": ["listField"],
                }
            ],
        )

    def does_not_accept_iterable_string_literal_as_a_list_value():
        assert _complete("Singular") == (
            {"listField": None},
            [
                {
                    "message": "Expected Iterable,"
                    " but did not find one for field 'Query.listField'.",
                    "locations": [(1, 3)],
                    "path": ["listField"],
                }
            ],
        )


def describe_execute_handles_list_nullability():
    async def _complete(list_field: Any, as_type: str) -> ExecutionResult:
        schema = build_schema(f"type Query {{ listField: {as_type} }}")
        document = parse("{ listField }")

        def execute_query(list_value: Any) -> Any:
            return execute(schema, document, Data(list_value))

        result = execute_query(list_field)
        assert isinstance(result, ExecutionResult)
        assert await execute_query(get_async(list_field)) == result
        if isinstance(list_field, list):
            assert await execute_query(list(map(get_async, list_field))) == result
            assert await execute_query(get_async(list_field)) == result

        return result

    @mark.asyncio
    async def contains_values():
        list_field = [1, 2]
        assert await _complete(list_field, "[Int]") == ({"listField": [1, 2]}, None)
        assert await _complete(list_field, "[Int]!") == ({"listField": [1, 2]}, None)
        assert await _complete(list_field, "[Int!]") == ({"listField": [1, 2]}, None)
        assert await _complete(list_field, "[Int!]!") == ({"listField": [1, 2]}, None)

    @mark.asyncio
    async def contains_null():
        list_field = [1, None, 2]
        errors = [
            {
                "message": "Cannot return null for non-nullable field Query.listField.",
                "locations": [(1, 3)],
                "path": ["listField", 1],
            }
        ]
        assert await _complete(list_field, "[Int]") == (
            {"listField": [1, None, 2]},
            None,
        )
        assert await _complete(list_field, "[Int]!") == (
            {"listField": [1, None, 2]},
            None,
        )
        assert await _complete(list_field, "[Int!]") == ({"listField": None}, errors)
        assert await _complete(list_field, "[Int!]!") == (None, errors)

    @mark.asyncio
    async def returns_null():
        list_field = None
        errors = [
            {
                "message": "Cannot return null for non-nullable field Query.listField.",
                "locations": [(1, 3)],
                "path": ["listField"],
            }
        ]
        assert await _complete(list_field, "[Int]") == ({"listField": None}, None)
        assert await _complete(list_field, "[Int]!") == (None, errors)
        assert await _complete(list_field, "[Int!]") == ({"listField": None}, None)
        assert await _complete(list_field, "[Int!]!") == (None, errors)

    @mark.asyncio
    async def contains_error():
        list_field = [1, RuntimeError("bad"), 2]
        errors = [
            {
                "message": "bad",
                "locations": [(1, 3)],
                "path": ["listField", 1],
            }
        ]
        assert await _complete(list_field, "[Int]") == (
            {"listField": [1, None, 2]},
            errors,
        )
        assert await _complete(list_field, "[Int]!") == (
            {"listField": [1, None, 2]},
            errors,
        )
        assert await _complete(list_field, "[Int!]") == (
            {"listField": None},
            errors,
        )
        assert await _complete(list_field, "[Int!]!") == (
            None,
            errors,
        )

    @mark.asyncio
    async def results_in_errors():
        list_field = RuntimeError("bad")
        errors = [
            {
                "message": "bad",
                "locations": [(1, 3)],
                "path": ["listField"],
            }
        ]
        assert await _complete(list_field, "[Int]") == (
            {"listField": None},
            errors,
        )
        assert await _complete(list_field, "[Int]!") == (
            None,
            errors,
        )
        assert await _complete(list_field, "[Int!]") == (
            {"listField": None},
            errors,
        )
        assert await _complete(list_field, "[Int!]!") == (
            None,
            errors,
        )


def describe_experimental_execute_accepts_async_iterables_as_list_value():
    async def _complete(list_field):
        result = execute(
            build_schema("type Query { listField: [String] }"),
            parse("{ listField }"),
            Data(list_field),
        )
        assert is_awaitable(result)
        result = cast(Awaitable, result)
        return await result

    @mark.asyncio
    async def accepts_an_async_generator_as_a_list_value():
        async def list_field():
            yield "one"
            yield 2
            yield True

        assert await _complete(list_field()) == (
            {"listField": ["one", "2", "true"]},
            None,
        )

    @mark.asyncio
    async def accepts_a_custom_async_iterable_as_a_list_value():
        class ListField:
            def __aiter__(self):
                self.last = "hello"
                return self

            async def __anext__(self):
                last = self.last
                if last == "stop":
                    raise StopAsyncIteration
                self.last = "world" if last == "hello" else "stop"
                return last

        assert await _complete(ListField()) == (
            {"listField": ["hello", "world"]},
            None,
        )



from functools import partial

from graphql.validation import FragmentsOnCompositeTypesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, FragmentsOnCompositeTypesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_fragments_on_composite_types():
    def object_is_valid_fragment_type():
        assert_valid(
            """
            fragment validFragment on Dog {
              barks
            }
            """
        )

    def interface_is_valid_fragment_type():
        assert_valid(
            """
            fragment validFragment on Pet {
              name
            }
            """
        )

    def object_is_valid_inline_fragment_type():
        assert_valid(
            """
            fragment validFragment on Pet {
              ... on Dog {
                barks
              }
            }
            """
        )

    def interface_is_valid_inline_fragment_type():
        assert_valid(
            """
            fragment validFragment on Mammal {
              ... on Canine {
                name
              }
            }
            """
        )

    def inline_fragment_without_type_is_valid():
        assert_valid(
            """
            fragment validFragment on Pet {
              ... {
                name
              }
            }
            """
        )

    def union_is_valid_fragment_type():
        assert_valid(
            """
            fragment validFragment on CatOrDog {
              __typename
            }
            """
        )

    def scalar_is_invalid_fragment_type():
        assert_errors(
            """
            fragment scalarFragment on Boolean {
              bad
            }
            """,
            [
                {
                    "message": "Fragment 'scalarFragment' cannot condition"
                    " on non composite type 'Boolean'.",
                    "locations": [(2, 40)],
                },
            ],
        )

    def enum_is_invalid_fragment_type():
        assert_errors(
            """
            fragment scalarFragment on FurColor {
              bad
            }
            """,
            [
                {
                    "message": "Fragment 'scalarFragment' cannot condition"
                    " on non composite type 'FurColor'.",
                    "locations": [(2, 40)],
                },
            ],
        )

    def input_object_is_invalid_fragment_type():
        assert_errors(
            """
            fragment inputFragment on ComplexInput {
              stringField
            }
            """,
            [
                {
                    "message": "Fragment 'inputFragment' cannot condition"
                    " on non composite type 'ComplexInput'.",
                    "locations": [(2, 39)],
                },
            ],
        )

    def scalar_is_invalid_inline_fragment_type():
        assert_errors(
            """
            fragment invalidFragment on Pet {
              ... on String {
                barks
              }
            }
            """,
            [
                {
                    "message": "Fragment cannot condition"
                    " on non composite type 'String'.",
                    "locations": [(3, 22)],
                }
            ],
        )



from typing import List

from graphql.error import GraphQLError
from graphql.language import parse, Source
from graphql.validation import validate

from .star_wars_schema import star_wars_schema


def validation_errors(query: str) -> List[GraphQLError]:
    """Helper function to test a query and the expected response."""
    source = Source(query, "StarWars.graphql")
    ast = parse(source)
    return validate(star_wars_schema, ast)


def describe_star_wars_validation_tests():
    def describe_basic_queries():
        def validates_a_complex_but_valid_query():
            query = """
                query NestedQueryWithFragment {
                  hero {
                    ...NameAndAppearances
                    friends {
                      ...NameAndAppearances
                      friends {
                        ...NameAndAppearances
                      }
                    }
                  }
                }

                fragment NameAndAppearances on Character {
                  name
                  appearsIn
                }
                """
            assert not validation_errors(query)

        def notes_that_non_existent_fields_are_invalid():
            query = """
                query HeroSpaceshipQuery {
                  hero {
                    favoriteSpaceship
                  }
                }
                """
            assert validation_errors(query)

        def requires_fields_on_object():
            query = """
                query HeroNoFieldsQuery {
                  hero
                }
                """
            assert validation_errors(query)

        def disallows_fields_on_scalars():
            query = """
                query HeroFieldsOnScalarQuery {
                  hero {
                    name {
                      firstCharacterOfName
                    }
                  }
                }
                """
            assert validation_errors(query)

        def disallows_object_fields_on_interfaces():
            query = """
                query DroidFieldOnCharacter {
                  hero {
                    name
                    primaryFunction
                  }
                }
                """
            assert validation_errors(query)

        def allows_object_fields_in_fragments():
            query = """
                query DroidFieldInFragment {
                  hero {
                    name
                    ...DroidFields
                  }
                }

                fragment DroidFields on Droid {
                  primaryFunction
                }
                """
            assert not validation_errors(query)

        def allows_object_fields_in_inline_fragments():
            query = """
                query DroidFieldInFragment {
                  hero {
                    name
                    ...DroidFields
                  }
                }

                fragment DroidFields on Droid {
                    primaryFunction
                }
                """
            assert not validation_errors(query)



from operator import attrgetter
from typing import Any

from ...error import GraphQLError
from ...language import OperationDefinitionNode
from ...pyutils import group_by
from . import ASTValidationRule

__all__ = ["UniqueVariableNamesRule"]


class UniqueVariableNamesRule(ASTValidationRule):
    """Unique variable names

    A GraphQL operation is only valid if all its variables are uniquely named.
    """

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_args: Any
    ) -> None:
        variable_definitions = node.variable_definitions

        seen_variable_definitions = group_by(
            variable_definitions, attrgetter("variable.name.value")
        )

        for variable_name, variable_nodes in seen_variable_definitions.items():
            if len(variable_nodes) > 1:
                self.report_error(
                    GraphQLError(
                        f"There can be only one variable named '${variable_name}'.",
                        [node.variable.name for node in variable_nodes],
                    )
                )



from collections import defaultdict
from typing import Any, Dict, List, Union, cast

from ...error import GraphQLError
from ...language import (
    DirectiveDefinitionNode,
    DirectiveNode,
    Node,
    SchemaDefinitionNode,
    SchemaExtensionNode,
    TypeDefinitionNode,
    TypeExtensionNode,
    is_type_definition_node,
    is_type_extension_node,
)
from ...type import specified_directives
from . import ASTValidationRule, SDLValidationContext, ValidationContext

__all__ = ["UniqueDirectivesPerLocationRule"]


class UniqueDirectivesPerLocationRule(ASTValidationRule):
    """Unique directive names per location

    A GraphQL document is only valid if all non-repeatable directives at a given
    location are uniquely named.

    See https://spec.graphql.org/draft/#sec-Directives-Are-Unique-Per-Location
    """

    context: Union[ValidationContext, SDLValidationContext]

    def __init__(self, context: Union[ValidationContext, SDLValidationContext]):
        super().__init__(context)
        unique_directive_map: Dict[str, bool] = {}

        schema = context.schema
        defined_directives = (
            schema.directives if schema else cast(List, specified_directives)
        )
        for directive in defined_directives:
            unique_directive_map[directive.name] = not directive.is_repeatable

        ast_definitions = context.document.definitions
        for def_ in ast_definitions:
            if isinstance(def_, DirectiveDefinitionNode):
                unique_directive_map[def_.name.value] = not def_.repeatable
        self.unique_directive_map = unique_directive_map

        self.schema_directives: Dict[str, DirectiveNode] = {}
        self.type_directives_map: Dict[str, Dict[str, DirectiveNode]] = defaultdict(
            dict
        )

    # Many different AST nodes may contain directives. Rather than listing them all,
    # just listen for entering any node, and check to see if it defines any directives.
    def enter(self, node: Node, *_args: Any) -> None:
        directives = getattr(node, "directives", None)
        if not directives:
            return
        directives = cast(List[DirectiveNode], directives)

        if isinstance(node, (SchemaDefinitionNode, SchemaExtensionNode)):
            seen_directives = self.schema_directives
        elif is_type_definition_node(node) or is_type_extension_node(node):
            node = cast(Union[TypeDefinitionNode, TypeExtensionNode], node)
            type_name = node.name.value
            seen_directives = self.type_directives_map[type_name]
        else:
            seen_directives = {}

        for directive in directives:
            directive_name = directive.name.value

            if self.unique_directive_map.get(directive_name):
                if directive_name in seen_directives:
                    self.report_error(
                        GraphQLError(
                            f"The directive '@{directive_name}'"
                            " can only be used once at this location.",
                            [seen_directives[directive_name], directive],
                        )
                    )
                else:
                    seen_directives[directive_name] = directive



from enum import Enum
from typing import Any, Collection, Dict, List, NamedTuple, Union, cast

from ..language import print_ast
from ..pyutils import inspect, Undefined
from ..type import (
    GraphQLEnumType,
    GraphQLField,
    GraphQLList,
    GraphQLNamedType,
    GraphQLNonNull,
    GraphQLInputType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLType,
    GraphQLUnionType,
    is_enum_type,
    is_input_object_type,
    is_interface_type,
    is_list_type,
    is_named_type,
    is_non_null_type,
    is_object_type,
    is_required_argument,
    is_required_input_field,
    is_scalar_type,
    is_specified_scalar_type,
    is_union_type,
)
from ..utilities.sort_value_node import sort_value_node
from .ast_from_value import ast_from_value

__all__ = [
    "BreakingChange",
    "BreakingChangeType",
    "DangerousChange",
    "DangerousChangeType",
    "find_breaking_changes",
    "find_dangerous_changes",
]


class BreakingChangeType(Enum):
    TYPE_REMOVED = 10
    TYPE_CHANGED_KIND = 11
    TYPE_REMOVED_FROM_UNION = 20
    VALUE_REMOVED_FROM_ENUM = 21
    REQUIRED_INPUT_FIELD_ADDED = 22
    IMPLEMENTED_INTERFACE_REMOVED = 23
    FIELD_REMOVED = 30
    FIELD_CHANGED_KIND = 31
    REQUIRED_ARG_ADDED = 40
    ARG_REMOVED = 41
    ARG_CHANGED_KIND = 42
    DIRECTIVE_REMOVED = 50
    DIRECTIVE_ARG_REMOVED = 51
    REQUIRED_DIRECTIVE_ARG_ADDED = 52
    DIRECTIVE_REPEATABLE_REMOVED = 53
    DIRECTIVE_LOCATION_REMOVED = 54


class DangerousChangeType(Enum):
    VALUE_ADDED_TO_ENUM = 60
    TYPE_ADDED_TO_UNION = 61
    OPTIONAL_INPUT_FIELD_ADDED = 62
    OPTIONAL_ARG_ADDED = 63
    IMPLEMENTED_INTERFACE_ADDED = 64
    ARG_DEFAULT_VALUE_CHANGE = 65


class BreakingChange(NamedTuple):
    type: BreakingChangeType
    description: str


class DangerousChange(NamedTuple):
    type: DangerousChangeType
    description: str


Change = Union[BreakingChange, DangerousChange]


def find_breaking_changes(
    old_schema: GraphQLSchema, new_schema: GraphQLSchema
) -> List[BreakingChange]:
    """Find breaking changes.

    Given two schemas, returns a list containing descriptions of all the types of
    breaking changes covered by the other functions down below.
    """
    return [
        change
        for change in find_schema_changes(old_schema, new_schema)
        if isinstance(change.type, BreakingChangeType)
    ]


def find_dangerous_changes(
    old_schema: GraphQLSchema, new_schema: GraphQLSchema
) -> List[DangerousChange]:
    """Find dangerous changes.

    Given two schemas, returns a list containing descriptions of all the types of
    potentially dangerous changes covered by the other functions down below.
    """
    return [
        change
        for change in find_schema_changes(old_schema, new_schema)
        if isinstance(change.type, DangerousChangeType)
    ]


def find_schema_changes(
    old_schema: GraphQLSchema, new_schema: GraphQLSchema
) -> List[Change]:
    return find_type_changes(old_schema, new_schema) + find_directive_changes(
        old_schema, new_schema
    )


def find_directive_changes(
    old_schema: GraphQLSchema, new_schema: GraphQLSchema
) -> List[Change]:
    schema_changes: List[Change] = []

    directives_diff = list_diff(old_schema.directives, new_schema.directives)

    for directive in directives_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.DIRECTIVE_REMOVED, f"{directive.name} was removed."
            )
        )

    for (old_directive, new_directive) in directives_diff.persisted:
        args_diff = dict_diff(old_directive.args, new_directive.args)

        for arg_name, new_arg in args_diff.added.items():
            if is_required_argument(new_arg):
                schema_changes.append(
                    BreakingChange(
                        BreakingChangeType.REQUIRED_DIRECTIVE_ARG_ADDED,
                        f"A required arg {arg_name} on directive"
                        f" {old_directive.name} was added.",
                    )
                )

        for arg_name in args_diff.removed:
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.DIRECTIVE_ARG_REMOVED,
                    f"{arg_name} was removed from {new_directive.name}.",
                )
            )

        if old_directive.is_repeatable and not new_directive.is_repeatable:
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.DIRECTIVE_REPEATABLE_REMOVED,
                    f"Repeatable flag was removed from {old_directive.name}.",
                )
            )

        for location in old_directive.locations:
            if location not in new_directive.locations:
                schema_changes.append(
                    BreakingChange(
                        BreakingChangeType.DIRECTIVE_LOCATION_REMOVED,
                        f"{location.name} was removed from {new_directive.name}.",
                    )
                )

    return schema_changes


def find_type_changes(
    old_schema: GraphQLSchema, new_schema: GraphQLSchema
) -> List[Change]:
    schema_changes: List[Change] = []
    types_diff = dict_diff(old_schema.type_map, new_schema.type_map)

    for type_name, old_type in types_diff.removed.items():
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.TYPE_REMOVED,
                f"Standard scalar {type_name} was removed"
                " because it is not referenced anymore."
                if is_specified_scalar_type(old_type)
                else f"{type_name} was removed.",
            )
        )

    for type_name, (old_type, new_type) in types_diff.persisted.items():
        if is_enum_type(old_type) and is_enum_type(new_type):
            schema_changes.extend(find_enum_type_changes(old_type, new_type))
        elif is_union_type(old_type) and is_union_type(new_type):
            schema_changes.extend(find_union_type_changes(old_type, new_type))
        elif is_input_object_type(old_type) and is_input_object_type(new_type):
            schema_changes.extend(find_input_object_type_changes(old_type, new_type))
        elif is_object_type(old_type) and is_object_type(new_type):
            schema_changes.extend(find_field_changes(old_type, new_type))
            schema_changes.extend(
                find_implemented_interfaces_changes(old_type, new_type)
            )
        elif is_interface_type(old_type) and is_interface_type(new_type):
            schema_changes.extend(find_field_changes(old_type, new_type))
            schema_changes.extend(
                find_implemented_interfaces_changes(old_type, new_type)
            )
        elif old_type.__class__ is not new_type.__class__:
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.TYPE_CHANGED_KIND,
                    f"{type_name} changed from {type_kind_name(old_type)}"
                    f" to {type_kind_name(new_type)}.",
                )
            )

    return schema_changes


def find_input_object_type_changes(
    old_type: Union[GraphQLObjectType, GraphQLInterfaceType],
    new_type: Union[GraphQLObjectType, GraphQLInterfaceType],
) -> List[Change]:
    schema_changes: List[Change] = []
    fields_diff = dict_diff(old_type.fields, new_type.fields)

    for field_name, new_field in fields_diff.added.items():
        if is_required_input_field(new_field):
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.REQUIRED_INPUT_FIELD_ADDED,
                    f"A required field {field_name} on"
                    f" input type {old_type.name} was added.",
                )
            )
        else:
            schema_changes.append(
                DangerousChange(
                    DangerousChangeType.OPTIONAL_INPUT_FIELD_ADDED,
                    f"An optional field {field_name} on"
                    f" input type {old_type.name} was added.",
                )
            )

    for field_name in fields_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.FIELD_REMOVED,
                f"{old_type.name}.{field_name} was removed.",
            )
        )

    for field_name, (old_field, new_field) in fields_diff.persisted.items():
        is_safe = is_change_safe_for_input_object_field_or_field_arg(
            old_field.type, new_field.type
        )
        if not is_safe:
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.FIELD_CHANGED_KIND,
                    f"{old_type.name}.{field_name} changed type"
                    f" from {old_field.type} to {new_field.type}.",
                )
            )

    return schema_changes


def find_union_type_changes(
    old_type: GraphQLUnionType, new_type: GraphQLUnionType
) -> List[Change]:
    schema_changes: List[Change] = []
    possible_types_diff = list_diff(old_type.types, new_type.types)

    for possible_type in possible_types_diff.added:
        schema_changes.append(
            DangerousChange(
                DangerousChangeType.TYPE_ADDED_TO_UNION,
                f"{possible_type.name} was added" f" to union type {old_type.name}.",
            )
        )

    for possible_type in possible_types_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.TYPE_REMOVED_FROM_UNION,
                f"{possible_type.name} was removed from union type {old_type.name}.",
            )
        )

    return schema_changes


def find_enum_type_changes(
    old_type: GraphQLEnumType, new_type: GraphQLEnumType
) -> List[Change]:
    schema_changes: List[Change] = []
    values_diff = dict_diff(old_type.values, new_type.values)

    for value_name in values_diff.added:
        schema_changes.append(
            DangerousChange(
                DangerousChangeType.VALUE_ADDED_TO_ENUM,
                f"{value_name} was added to enum type {old_type.name}.",
            )
        )

    for value_name in values_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.VALUE_REMOVED_FROM_ENUM,
                f"{value_name} was removed from enum type {old_type.name}.",
            )
        )

    return schema_changes


def find_implemented_interfaces_changes(
    old_type: Union[GraphQLObjectType, GraphQLInterfaceType],
    new_type: Union[GraphQLObjectType, GraphQLInterfaceType],
) -> List[Change]:
    schema_changes: List[Change] = []
    interfaces_diff = list_diff(old_type.interfaces, new_type.interfaces)

    for interface in interfaces_diff.added:
        schema_changes.append(
            DangerousChange(
                DangerousChangeType.IMPLEMENTED_INTERFACE_ADDED,
                f"{interface.name} added to interfaces implemented by {old_type.name}.",
            )
        )

    for interface in interfaces_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.IMPLEMENTED_INTERFACE_REMOVED,
                f"{old_type.name} no longer implements interface {interface.name}.",
            )
        )

    return schema_changes


def find_field_changes(
    old_type: Union[GraphQLObjectType, GraphQLInterfaceType],
    new_type: Union[GraphQLObjectType, GraphQLInterfaceType],
) -> List[Change]:
    schema_changes: List[Change] = []
    fields_diff = dict_diff(old_type.fields, new_type.fields)

    for field_name in fields_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.FIELD_REMOVED,
                f"{old_type.name}.{field_name} was removed.",
            )
        )

    for field_name, (old_field, new_field) in fields_diff.persisted.items():
        schema_changes.extend(
            find_arg_changes(old_type, field_name, old_field, new_field)
        )
        is_safe = is_change_safe_for_object_or_interface_field(
            old_field.type, new_field.type
        )
        if not is_safe:
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.FIELD_CHANGED_KIND,
                    f"{old_type.name}.{field_name} changed type"
                    f" from {old_field.type} to {new_field.type}.",
                )
            )

    return schema_changes


def find_arg_changes(
    old_type: Union[GraphQLObjectType, GraphQLInterfaceType],
    field_name: str,
    old_field: GraphQLField,
    new_field: GraphQLField,
) -> List[Change]:
    schema_changes: List[Change] = []
    args_diff = dict_diff(old_field.args, new_field.args)

    for arg_name in args_diff.removed:
        schema_changes.append(
            BreakingChange(
                BreakingChangeType.ARG_REMOVED,
                f"{old_type.name}.{field_name} arg" f" {arg_name} was removed.",
            )
        )

    for arg_name, (old_arg, new_arg) in args_diff.persisted.items():
        is_safe = is_change_safe_for_input_object_field_or_field_arg(
            old_arg.type, new_arg.type
        )
        if not is_safe:
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.ARG_CHANGED_KIND,
                    f"{old_type.name}.{field_name} arg"
                    f" {arg_name} has changed type from"
                    f" {old_arg.type} to {new_arg.type}.",
                )
            )
        elif old_arg.default_value is not Undefined:
            if new_arg.default_value is Undefined:
                schema_changes.append(
                    DangerousChange(
                        DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                        f"{old_type.name}.{field_name} arg"
                        f" {arg_name} defaultValue was removed.",
                    )
                )
            else:
                # Since we are looking only for client's observable changes we should
                # compare default values in the same representation as they are
                # represented inside introspection.
                old_value_str = stringify_value(old_arg.default_value, old_arg.type)
                new_value_str = stringify_value(new_arg.default_value, new_arg.type)

                if old_value_str != new_value_str:
                    schema_changes.append(
                        DangerousChange(
                            DangerousChangeType.ARG_DEFAULT_VALUE_CHANGE,
                            f"{old_type.name}.{field_name} arg"
                            f" {arg_name} has changed defaultValue"
                            f" from {old_value_str} to {new_value_str}.",
                        )
                    )

    for arg_name, new_arg in args_diff.added.items():
        if is_required_argument(new_arg):
            schema_changes.append(
                BreakingChange(
                    BreakingChangeType.REQUIRED_ARG_ADDED,
                    f"A required arg {arg_name} on"
                    f" {old_type.name}.{field_name} was added.",
                )
            )
        else:
            schema_changes.append(
                DangerousChange(
                    DangerousChangeType.OPTIONAL_ARG_ADDED,
                    f"An optional arg {arg_name} on"
                    f" {old_type.name}.{field_name} was added.",
                )
            )

    return schema_changes


def is_change_safe_for_object_or_interface_field(
    old_type: GraphQLType, new_type: GraphQLType
) -> bool:
    if is_list_type(old_type):
        return (
            # if they're both lists, make sure underlying types are compatible
            is_list_type(new_type)
            and is_change_safe_for_object_or_interface_field(
                cast(GraphQLList, old_type).of_type, cast(GraphQLList, new_type).of_type
            )
        ) or (
            # moving from nullable to non-null of same underlying type is safe
            is_non_null_type(new_type)
            and is_change_safe_for_object_or_interface_field(
                old_type, cast(GraphQLNonNull, new_type).of_type
            )
        )

    if is_non_null_type(old_type):
        # if they're both non-null, make sure underlying types are compatible
        return is_non_null_type(
            new_type
        ) and is_change_safe_for_object_or_interface_field(
            cast(GraphQLNonNull, old_type).of_type,
            cast(GraphQLNonNull, new_type).of_type,
        )

    return (
        # if they're both named types, see if their names are equivalent
        is_named_type(new_type)
        and cast(GraphQLNamedType, old_type).name
        == cast(GraphQLNamedType, new_type).name
    ) or (
        # moving from nullable to non-null of same underlying type is safe
        is_non_null_type(new_type)
        and is_change_safe_for_object_or_interface_field(
            old_type, cast(GraphQLNonNull, new_type).of_type
        )
    )


def is_change_safe_for_input_object_field_or_field_arg(
    old_type: GraphQLType, new_type: GraphQLType
) -> bool:
    if is_list_type(old_type):

        return is_list_type(
            # if they're both lists, make sure underlying types are compatible
            new_type
        ) and is_change_safe_for_input_object_field_or_field_arg(
            cast(GraphQLList, old_type).of_type, cast(GraphQLList, new_type).of_type
        )

    if is_non_null_type(old_type):
        return (
            # if they're both non-null, make sure the underlying types are compatible
            is_non_null_type(new_type)
            and is_change_safe_for_input_object_field_or_field_arg(
                cast(GraphQLNonNull, old_type).of_type,
                cast(GraphQLNonNull, new_type).of_type,
            )
        ) or (
            # moving from non-null to nullable of same underlying type is safe
            not is_non_null_type(new_type)
            and is_change_safe_for_input_object_field_or_field_arg(
                cast(GraphQLNonNull, old_type).of_type, new_type
            )
        )

    return (
        # if they're both named types, see if their names are equivalent
        is_named_type(new_type)
        and cast(GraphQLNamedType, old_type).name
        == cast(GraphQLNamedType, new_type).name
    )


def type_kind_name(type_: GraphQLNamedType) -> str:
    if is_scalar_type(type_):
        return "a Scalar type"
    if is_object_type(type_):
        return "an Object type"
    if is_interface_type(type_):
        return "an Interface type"
    if is_union_type(type_):
        return "a Union type"
    if is_enum_type(type_):
        return "an Enum type"
    if is_input_object_type(type_):
        return "an Input type"

    # Not reachable. All possible output types have been considered.
    raise TypeError(f"Unexpected type {inspect(type)}")


def stringify_value(value: Any, type_: GraphQLInputType) -> str:
    ast = ast_from_value(value, type_)
    if ast is None:  # pragma: no cover
        raise TypeError(f"Invalid value: {inspect(value)}")
    return print_ast(sort_value_node(ast))


class ListDiff(NamedTuple):
    """Tuple with added, removed and persisted list items."""

    added: List
    removed: List
    persisted: List


def list_diff(old_list: Collection, new_list: Collection) -> ListDiff:
    """Get differences between two lists of named items."""
    added = []
    persisted = []
    removed = []

    old_set = {item.name for item in old_list}
    new_map = {item.name: item for item in new_list}

    for old_item in old_list:
        new_item = new_map.get(old_item.name)
        if new_item:
            persisted.append([old_item, new_item])
        else:
            removed.append(old_item)

    for new_item in new_list:
        if new_item.name not in old_set:
            added.append(new_item)

    return ListDiff(added, removed, persisted)


class DictDiff(NamedTuple):
    """Tuple with added, removed and persisted dict entries."""

    added: Dict
    removed: Dict
    persisted: Dict


def dict_diff(old_dict: Dict, new_dict: Dict) -> DictDiff:
    """Get differences between two dicts."""
    added = {}
    removed = {}
    persisted = {}

    for old_name, old_item in old_dict.items():
        new_item = new_dict.get(old_name)
        if new_item:
            persisted[old_name] = [old_item, new_item]
        else:
            removed[old_name] = old_item

    for new_name, new_item in new_dict.items():
        if new_name not in old_dict:
            added[new_name] = new_item

    return DictDiff(added, removed, persisted)



from string import ascii_letters as letters, digits, punctuation

from graphql.language.character_classes import (
    is_digit,
    is_letter,
    is_name_start,
    is_name_continue,
)

non_ascii = "¯＿±¹²³½£ºµÄäÖöØø×〇᧐〸αΑωΩ"


def describe_digit():
    def accepts_digits():
        assert all(is_digit(char) for char in digits)

    def rejects_letters():
        assert not any(is_digit(char) for char in letters)

    def rejects_underscore():
        assert not is_digit("_")

    def rejects_punctuation():
        assert not any(is_digit(char) for char in punctuation)

    def rejects_non_ascii():
        assert not any(is_digit(char) for char in non_ascii)

    def rejects_empty_string():
        assert not is_digit("")


def describe_letter():
    def accepts_letters():
        assert all(is_letter(char) for char in letters)

    def rejects_digits():
        assert not any(is_letter(char) for char in digits)

    def rejects_underscore():
        assert not is_letter("_")

    def rejects_punctuation():
        assert not any(is_letter(char) for char in punctuation)

    def rejects_non_ascii():
        assert not any(is_letter(char) for char in non_ascii)

    def rejects_empty_string():
        assert not is_letter("")


def describe_name_start():
    def accepts_letters():
        assert all(is_name_start(char) for char in letters)

    def accepts_underscore():
        assert is_name_start("_")

    def rejects_digits():
        assert not any(is_name_start(char) for char in digits)

    def rejects_punctuation():
        assert not any(is_name_start(char) for char in punctuation if char != "_")

    def rejects_non_ascii():
        assert not any(is_name_start(char) for char in non_ascii)

    def rejects_empty_string():
        assert not is_name_start("")


def describe_name_continue():
    def accepts_letters():
        assert all(is_name_continue(char) for char in letters)

    def accepts_digits():
        assert all(is_name_continue(char) for char in digits)

    def accepts_underscore():
        assert is_name_continue("_")

    def rejects_punctuation():
        assert not any(is_name_continue(char) for char in punctuation if char != "_")

    def rejects_non_ascii():
        assert not any(is_name_continue(char) for char in non_ascii)

    def rejects_empty_string():
        assert not is_name_continue("")



from graphql.pyutils import group_by


def describe_group_by():
    def does_accept_an_empty_list():
        def key_fn(_x: str) -> str:
            raise TypeError("Unexpected call of key function.")

        assert group_by([], key_fn) == {}

    def does_not_change_order():
        def key_fn(_x: int) -> str:
            return "all"

        assert group_by([3, 1, 5, 4, 2, 6], key_fn) == {
            "all": [3, 1, 5, 4, 2, 6],
        }

    def can_group_by_odd_and_even():
        def key_fn(x: int) -> str:
            return "odd" if x % 2 else "even"

        assert group_by([3, 1, 5, 4, 2, 6], key_fn) == {
            "odd": [3, 1, 5],
            "even": [4, 2, 6],
        }

    def can_group_by_string_length():
        def key_fn(s: str) -> int:
            return len(s)

        assert group_by(
            [
                "alpha",
                "beta",
                "gamma",
                "delta",
                "epsilon",
                "zeta",
                "eta",
                "iota",
                "kapp",
                "lambda",
                "my",
                "ny",
                "omikron",
            ],
            key_fn,
        ) == {
            2: ["my", "ny"],
            3: ["eta"],
            4: ["beta", "zeta", "iota", "kapp"],
            5: ["alpha", "gamma", "delta"],
            6: ["lambda"],
            7: ["epsilon", "omikron"],
        }



from pytest import mark

from graphql import graphql, graphql_sync

from .star_wars_schema import star_wars_schema as schema


def describe_star_wars_query_tests():
    def describe_basic_queries():
        @mark.asyncio
        async def correctly_identifies_r2_d2_as_hero_of_the_star_wars_saga():
            source = """
                query HeroNameQuery {
                  hero {
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == ({"hero": {"name": "R2-D2"}}, None)

        @mark.asyncio
        async def accepts_positional_arguments_to_graphql():
            source = """
                query HeroNameQuery {
                  hero {
                    name
                  }
                }
                """
            result = await graphql(schema, source)
            assert result == ({"hero": {"name": "R2-D2"}}, None)

            sync_result = graphql_sync(schema, source)
            assert sync_result == result

        @mark.asyncio
        async def allows_us_to_query_for_the_id_and_friends_of_r2_d2():
            source = """
                query HeroNameAndFriendsQuery {
                  hero {
                    id
                    name
                    friends {
                      name
                    }
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {
                    "hero": {
                        "id": "2001",
                        "name": "R2-D2",
                        "friends": [
                            {"name": "Luke Skywalker"},
                            {"name": "Han Solo"},
                            {"name": "Leia Organa"},
                        ],
                    }
                },
                None,
            )

    def describe_nested_queries():
        @mark.asyncio
        async def allows_us_to_query_for_the_friends_of_friends_of_r2_d2():
            source = """
                query NestedQuery {
                  hero {
                    name
                    friends {
                      name
                      appearsIn
                      friends {
                        name
                      }
                    }
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {
                    "hero": {
                        "name": "R2-D2",
                        "friends": [
                            {
                                "name": "Luke Skywalker",
                                "appearsIn": ["NEW_HOPE", "EMPIRE", "JEDI"],
                                "friends": [
                                    {"name": "Han Solo"},
                                    {"name": "Leia Organa"},
                                    {"name": "C-3PO"},
                                    {"name": "R2-D2"},
                                ],
                            },
                            {
                                "name": "Han Solo",
                                "appearsIn": ["NEW_HOPE", "EMPIRE", "JEDI"],
                                "friends": [
                                    {"name": "Luke Skywalker"},
                                    {"name": "Leia Organa"},
                                    {"name": "R2-D2"},
                                ],
                            },
                            {
                                "name": "Leia Organa",
                                "appearsIn": ["NEW_HOPE", "EMPIRE", "JEDI"],
                                "friends": [
                                    {"name": "Luke Skywalker"},
                                    {"name": "Han Solo"},
                                    {"name": "C-3PO"},
                                    {"name": "R2-D2"},
                                ],
                            },
                        ],
                    }
                },
                None,
            )

    def describe_using_ids_and_query_parameters_to_refetch_objects():
        @mark.asyncio
        async def allows_us_to_query_for_r2_d2_directly_using_his_id():
            source = """
                query {
                  droid(id: "2001") {
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == ({"droid": {"name": "R2-D2"}}, None)

        @mark.asyncio
        async def allows_us_to_query_characters_directly_using_their_id():
            source = """
                query FetchLukeAndC3POQuery {
                  human(id: "1000") {
                    name
                  }
                  droid(id: "2000") {
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {"human": {"name": "Luke Skywalker"}, "droid": {"name": "C-3PO"}},
                None,
            )

        @mark.asyncio
        async def allows_creating_a_generic_query_to_fetch_luke_using_his_id():
            source = """
                query FetchSomeIDQuery($someId: String!) {
                  human(id: $someId) {
                    name
                  }
                }
                """
            variable_values = {"someId": "1000"}
            result = await graphql(
                schema=schema, source=source, variable_values=variable_values
            )
            assert result == ({"human": {"name": "Luke Skywalker"}}, None)

        @mark.asyncio
        async def allows_creating_a_generic_query_to_fetch_han_using_his_id():
            source = """
                query FetchSomeIDQuery($someId: String!) {
                  human(id: $someId) {
                    name
                  }
                }
                """
            variable_values = {"someId": "1002"}
            result = await graphql(
                schema=schema, source=source, variable_values=variable_values
            )
            assert result == ({"human": {"name": "Han Solo"}}, None)

        @mark.asyncio
        async def generic_query_that_gets_null_back_when_passed_invalid_id():
            source = """
                query humanQuery($id: String!) {
                  human(id: $id) {
                    name
                  }
                }
                """
            variable_values = {"id": "not a valid id"}
            result = await graphql(
                schema=schema, source=source, variable_values=variable_values
            )
            assert result == ({"human": None}, None)

    def describe_using_aliases_to_change_the_key_in_the_response():
        @mark.asyncio
        async def allows_us_to_query_for_luke_changing_his_key_with_an_alias():
            source = """
                query FetchLukeAliased {
                  luke: human(id: "1000") {
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == ({"luke": {"name": "Luke Skywalker"}}, None)

        @mark.asyncio
        async def query_for_luke_and_leia_using_two_root_fields_and_an_alias():
            source = """
                query FetchLukeAndLeiaAliased {
                  luke: human(id: "1000") {
                    name
                  }
                  leia: human(id: "1003") {
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {"luke": {"name": "Luke Skywalker"}, "leia": {"name": "Leia Organa"}},
                None,
            )

    def describe_uses_fragments_to_express_more_complex_queries():
        @mark.asyncio
        async def allows_us_to_query_using_duplicated_content():
            source = """
                query DuplicateFields {
                  luke: human(id: "1000") {
                    name
                    homePlanet
                  }
                  leia: human(id: "1003") {
                    name
                    homePlanet
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {
                    "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
                    "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
                },
                None,
            )

        @mark.asyncio
        async def allows_us_to_use_a_fragment_to_avoid_duplicating_content():
            source = """
                query UseFragment {
                  luke: human(id: "1000") {
                    ...HumanFragment
                  }
                  leia: human(id: "1003") {
                    ...HumanFragment
                  }
                }
                fragment HumanFragment on Human {
                  name
                  homePlanet
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {
                    "luke": {"name": "Luke Skywalker", "homePlanet": "Tatooine"},
                    "leia": {"name": "Leia Organa", "homePlanet": "Alderaan"},
                },
                None,
            )

    def describe_using_typename_to_find_the_type_of_an_object():
        @mark.asyncio
        async def allows_us_to_verify_that_r2_d2_is_a_droid():
            source = """
                query CheckTypeOfR2 {
                  hero {
                    __typename
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == ({"hero": {"__typename": "Droid", "name": "R2-D2"}}, None)

        @mark.asyncio
        async def allows_us_to_verify_that_luke_is_a_human():
            source = """
                query CheckTypeOfLuke {
                  hero(episode: EMPIRE) {
                    __typename
                    name
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {"hero": {"__typename": "Human", "name": "Luke Skywalker"}},
                None,
            )

    def describe_reporting_errors_raised_in_resolvers():
        @mark.asyncio
        async def correctly_reports_error_on_accessing_secret_backstory():
            source = """
                query HeroNameQuery {
                  hero {
                    name
                    secretBackstory
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {"hero": {"name": "R2-D2", "secretBackstory": None}},
                [
                    {
                        "message": "secretBackstory is secret.",
                        "locations": [(5, 21)],
                        "path": ["hero", "secretBackstory"],
                    }
                ],
            )

        @mark.asyncio
        async def correctly_reports_error_on_accessing_backstory_in_a_list():
            source = """
                query HeroNameQuery {
                  hero {
                    name
                    friends {
                      name
                      secretBackstory
                    }
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {
                    "hero": {
                        "name": "R2-D2",
                        "friends": [
                            {"name": "Luke Skywalker", "secretBackstory": None},
                            {"name": "Han Solo", "secretBackstory": None},
                            {"name": "Leia Organa", "secretBackstory": None},
                        ],
                    }
                },
                [
                    {
                        "message": "secretBackstory is secret.",
                        "locations": [(7, 23)],
                        "path": ["hero", "friends", 0, "secretBackstory"],
                    },
                    {
                        "message": "secretBackstory is secret.",
                        "locations": [(7, 23)],
                        "path": ["hero", "friends", 1, "secretBackstory"],
                    },
                    {
                        "message": "secretBackstory is secret.",
                        "locations": [(7, 23)],
                        "path": ["hero", "friends", 2, "secretBackstory"],
                    },
                ],
            )

        @mark.asyncio
        async def correctly_reports_error_on_accessing_through_an_alias():
            source = """
                query HeroNameQuery {
                  mainHero: hero {
                    name
                    story: secretBackstory
                  }
                }
                """
            result = await graphql(schema=schema, source=source)
            assert result == (
                {"mainHero": {"name": "R2-D2", "story": None}},
                [
                    {
                        "message": "secretBackstory is secret.",
                        "locations": [(5, 21)],
                        "path": ["mainHero", "story"],
                    }
                ],
            )



from typing import Any, Dict, List

from ...error import GraphQLError
from ...language import NameNode, ObjectFieldNode
from . import ASTValidationContext, ASTValidationRule

__all__ = ["UniqueInputFieldNamesRule"]


class UniqueInputFieldNamesRule(ASTValidationRule):
    """Unique input field names

    A GraphQL input object value is only valid if all supplied fields are uniquely
    named.

    See https://spec.graphql.org/draft/#sec-Input-Object-Field-Uniqueness
    """

    def __init__(self, context: ASTValidationContext):
        super().__init__(context)
        self.known_names_stack: List[Dict[str, NameNode]] = []
        self.known_names: Dict[str, NameNode] = {}

    def enter_object_value(self, *_args: Any) -> None:
        self.known_names_stack.append(self.known_names)
        self.known_names = {}

    def leave_object_value(self, *_args: Any) -> None:
        self.known_names = self.known_names_stack.pop()

    def enter_object_field(self, node: ObjectFieldNode, *_args: Any) -> None:
        known_names = self.known_names
        field_name = node.name.value
        if field_name in known_names:
            self.report_error(
                GraphQLError(
                    f"There can be only one input field named '{field_name}'.",
                    [known_names[field_name], node.name],
                )
            )
        else:
            known_names[field_name] = node.name



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import SingleFieldSubscriptionsRule

from .harness import assert_validation_errors

schema = build_schema(
    """
    type Message {
      body: String
      sender: String
    }

    type SubscriptionRoot {
      importantEmails: [String]
      notImportantEmails: [String]
      moreImportantEmails: [String]
      spamEmails: [String]
      deletedEmails: [String]
      newMessage: Message
    }

    type QueryRoot {
      dummy: String
    }

    schema {
      query: QueryRoot
      subscription: SubscriptionRoot
    }
    """
)

assert_errors = partial(
    assert_validation_errors, SingleFieldSubscriptionsRule, schema=schema
)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_subscriptions_with_single_field():
    def valid_subscription():
        assert_valid(
            """
            subscription ImportantEmails {
              importantEmails
            }
            """
        )

    def valid_subscription_with_fragment():
        assert_valid(
            """
            subscription sub {
              ...newMessageFields
            }

            fragment newMessageFields on SubscriptionRoot {
              newMessage {
                body
                sender
              }
            }
            """
        )

    def valid_subscription_with_fragment_and_field():
        assert_valid(
            """
            subscription sub {
              newMessage {
                body
              }
              ...newMessageFields
            }

            fragment newMessageFields on SubscriptionRoot {
              newMessage {
                body
                sender
              }
            }
            """
        )

    def fails_with_more_than_one_root_field():
        assert_errors(
            """
            subscription ImportantEmails {
              importantEmails
              notImportantEmails
            }
            """,
            [
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must select only one top level field.",
                    "locations": [(4, 15)],
                }
            ],
        )

    def fails_with_more_than_one_root_field_including_introspection():
        assert_errors(
            """
            subscription ImportantEmails {
              importantEmails
              __typename
            }
            """,
            [
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must select only one top level field.",
                    "locations": [(4, 15)],
                },
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must not select an introspection top level field.",
                    "locations": [(4, 15)],
                },
            ],
        )

    def fails_with_more_than_one_root_field_including_aliased_introspection():
        assert_errors(
            """
            subscription ImportantEmails {
              importantEmails
              ...Introspection
            }
            fragment Introspection on SubscriptionRoot {
              typename: __typename
            }
            """,
            [
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must select only one top level field.",
                    "locations": [(7, 15)],
                },
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must not select an introspection top level field.",
                    "locations": [(7, 15)],
                },
            ],
        )

    def fails_with_many_more_than_one_root_field():
        assert_errors(
            """
            subscription ImportantEmails {
              importantEmails
              notImportantEmails
              spamEmails
            }
            """,
            [
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must select only one top level field.",
                    "locations": [(4, 15), (5, 15)],
                }
            ],
        )

    def fails_with_more_than_one_root_field_via_fragments():
        assert_errors(
            """
            subscription ImportantEmails {
              importantEmails
              ... {
                more: moreImportantEmails
              }
              ...NotImportantEmails
            }
            fragment NotImportantEmails on SubscriptionRoot {
              notImportantEmails
              deleted: deletedEmails
              ...SpamEmails
            }
            fragment SpamEmails on SubscriptionRoot {
              spamEmails
            }
            """,
            [
                {
                    "message": "Subscription 'ImportantEmails'"
                    " must select only one top level field.",
                    "locations": [(5, 17), (10, 15), (11, 15), (15, 15)],
                },
            ],
        )

    def does_not_infinite_loop_on_recursive_fragments():
        assert_errors(
            """
            subscription NoInfiniteLoop {
              ...A
            }
            fragment A on SubscriptionRoot {
              ...A
            }
            """,
            [],
        )

    def fails_with_more_than_one_root_field_via_fragments_anonymous():
        assert_errors(
            """
            subscription {
              importantEmails
              ... {
                more: moreImportantEmails
                ...NotImportantEmails
              }
              ...NotImportantEmails
            }
            fragment NotImportantEmails on SubscriptionRoot {
              notImportantEmails
              deleted: deletedEmails
              ... {
                ... {
                  archivedEmails
                }
              }
              ...SpamEmails
            }
            fragment SpamEmails on SubscriptionRoot {
              spamEmails
              ...NonExistentFragment
            }
            """,
            [
                {
                    "message": "Anonymous Subscription"
                    " must select only one top level field.",
                    "locations": [(5, 17), (11, 15), (12, 15), (15, 19), (21, 15)],
                },
            ],
        )

    def fails_with_more_than_one_root_field_in_anonymous_subscriptions():
        assert_errors(
            """
            subscription {
              importantEmails
              notImportantEmails
            }
            """,
            [
                {
                    "message": "Anonymous Subscription"
                    " must select only one top level field.",
                    "locations": [(4, 15)],
                }
            ],
        )

    def fails_with_introspection_field():
        assert_errors(
            """
            subscription ImportantEmails {
              __typename
            }
            """,
            [
                {
                    "message": "Subscription 'ImportantEmails' must not"
                    " select an introspection top level field.",
                    "locations": [(3, 15)],
                }
            ],
        )

    def fails_with_introspection_field_in_anonymous_subscription():
        assert_errors(
            """
            subscription {
              __typename
            }
            """,
            [
                {
                    "message": "Anonymous Subscription must not"
                    " select an introspection top level field.",
                    "locations": [(3, 15)],
                }
            ],
        )

    def skips_if_not_subscription_type():
        empty_schema = build_schema(
            """
            type Query {
              dummy: String
            }
            """
        )
        assert_errors(
            """
            subscription {
              __typename
            }
            """,
            [],
            schema=empty_schema,
        )



from functools import partial

from graphql.validation import KnownFragmentNamesRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, KnownFragmentNamesRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_known_fragment_names():
    def known_fragment_names_are_valid():
        assert_valid(
            """
            {
              human(id: 4) {
                ...HumanFields1
                  ... on Human {
                    ...HumanFields2
                  }
                  ... {
                    name
                  }
              }
            }
            fragment HumanFields1 on Human {
              name
              ...HumanFields3
            }
            fragment HumanFields2 on Human {
              name
            }
            fragment HumanFields3 on Human {
              name
            }
            """
        )

    def unknown_fragment_names_are_invalid():
        assert_errors(
            """
            {
              human(id: 4) {
                ...UnknownFragment1
                ... on Human {
                  ...UnknownFragment2
                }
              }
            }
            fragment HumanFields on Human {
              name
              ...UnknownFragment3
            }
            """,
            [
                {
                    "message": "Unknown fragment 'UnknownFragment1'.",
                    "locations": [(4, 20)],
                },
                {
                    "message": "Unknown fragment 'UnknownFragment2'.",
                    "locations": [(6, 22)],
                },
                {
                    "message": "Unknown fragment 'UnknownFragment3'.",
                    "locations": [(12, 18)],
                },
            ],
        )



from typing import Any

from ...error import GraphQLError
from ...language import FragmentSpreadNode
from . import ValidationRule

__all__ = ["KnownFragmentNamesRule"]


class KnownFragmentNamesRule(ValidationRule):
    """Known fragment names

    A GraphQL document is only valid if all ``...Fragment`` fragment spreads refer to
    fragments defined in the same document.

    See https://spec.graphql.org/draft/#sec-Fragment-spread-target-defined
    """

    def enter_fragment_spread(self, node: FragmentSpreadNode, *_args: Any) -> None:
        fragment_name = node.name.value
        fragment = self.context.get_fragment(fragment_name)
        if not fragment:
            self.report_error(
                GraphQLError(f"Unknown fragment '{fragment_name}'.", node.name)
            )



# uses code from https://github.com/daveoncode/python-string-utils

import re

__all__ = ["camel_to_snake", "snake_to_camel"]

_re_camel_to_snake = re.compile(r"([a-z]|[A-Z0-9]+)(?=[A-Z])")
_re_snake_to_camel = re.compile(r"(_)([a-z\d])")


def camel_to_snake(s: str) -> str:
    """Convert from CamelCase to snake_case"""
    return _re_camel_to_snake.sub(r"\1_", s).lower()


def snake_to_camel(s: str, upper: bool = True) -> str:
    """Convert from snake_case to CamelCase

    If upper is set, then convert to upper CamelCase, otherwise the first character
    keeps its case.
    """
    s = _re_snake_to_camel.sub(lambda m: m.group(2).upper(), s)
    if upper:
        s = s[:1].upper() + s[1:]
    return s
